"""
AMS command-line interface and argument parsers.
"""
import logging
import argparse
import importlib
import platform
import sys
from time import strftime

from andes.shared import NCPUS_PHYSICAL

from ams.main import config_logger
from ams.utils.paths import get_log_dir
from ams.routines import routine_cli

logger = logging.getLogger(__name__)

command_aliases = {
    'selftest': ['st'],
}


def create_parser():
    """
    Create a parser for the command-line interface.

    Returns
    -------
    argparse.ArgumentParser
        Parser with all AMS options

    Notes
    -----
    Revised from the ANDES project (https://github.com/CURENT/andes).
    Original author: Hantao Cui
    License: GPL3
    """

    parser = argparse.ArgumentParser()

    parser.add_argument(
        '-v', '--verbose',
        help='Verbosity level in 10-DEBUG, 20-INFO, 30-WARNING, '
             'or 40-ERROR.',
        type=int, default=20, choices=(1, 10, 20, 30, 40))

    sub_parsers = parser.add_subparsers(dest='command', help='[run] run simulation routine; '
                                                             '[doc] quick documentation; '
                                                             '[misc] misc. functions; '
                                                             '[selftest] run self test; '
                                        )

    run = sub_parsers.add_parser('run')
    run.add_argument('filename', help='Case file name. Power flow is calculated by default.', nargs='*')
    run.add_argument('-r', '--routine', nargs='*', default=('pflow', ),
                     action='store', help='Simulation routine(s). Single routine or multiple separated with '
                                          'space. Run PFlow by default.',
                     choices=list(routine_cli.keys()))
    run.add_argument('-p', '--input-path', help='Path to case files', type=str, default='')
    run.add_argument('-a', '--addfile', help='Additional files used by some formats.')
    run.add_argument('-o', '--output-path', help='Output path prefix', type=str, default='')
    run.add_argument('-n', '--no-output', help='Force no output of any kind', action='store_true')
    run.add_argument('--ncpu', help='Number of parallel processes', type=int, default=NCPUS_PHYSICAL)
    run.add_argument('-c', '--convert', help='Convert to format.', type=str, default='', nargs='?')
    run.add_argument('-b', '--add-book', help='Add a template workbook for the specified model.', type=str)
    run.add_argument('--convert-all', help='Convert to format with all templates.', type=str, default='',
                     nargs='?')
    run.add_argument('--profile', action='store_true', help='Enable Python cProfiler')
    run.add_argument('-s', '--shell', action='store_true', help='Start in IPython shell')
    run.add_argument('--no-preamble', action='store_true', help='Hide preamble')
    run.add_argument('--no-pbar', action='store_true', help='Hide progress bar for time-domain')
    run.add_argument('--pool', action='store_true', help='Start multiprocess with Pool '
                                                         'and return a list of Systems')
    run.add_argument('--from-csv', help='Use data from a CSV file instead of from simulation')
    run.add_argument('-O', '--config-option',
                     help='Set configuration option specificied by '
                          'NAME.FIELD=VALUE with no space. For example, "TDS.tf=2"',
                     type=str, default='', nargs='*')

    doc = sub_parsers.add_parser('doc')
    doc.add_argument('--list', '-l', help='List supported routines', action='store_true',
                     dest='list_supported')

    misc = sub_parsers.add_parser('misc')
    config_exclusive = misc.add_mutually_exclusive_group()
    config_exclusive.add_argument('--edit-config', help='Quick edit of the config file',
                                  default='', nargs='?', type=str)
    config_exclusive.add_argument('--save-config', help='save configuration to file name',
                                  nargs='?', type=str, default='')
    misc.add_argument('--license', action='store_true', help='Display software license', dest='show_license')
    misc.add_argument('-C', '--clean', help='Clean output files', action='store_true')
    misc.add_argument('-r', '--recursive', help='Recursively clean outputs (combined useage with --clean)',
                      action='store_true')
    misc.add_argument('--version', action='store_true', help='Display version information')

    sub_parsers.add_parser('selftest', aliases=command_aliases['selftest'])  # NOQA

    return parser


def preamble():
    """
    Log the AMS command-line preamble at the `logging.INFO` level
    """
    from ams import __version__ as version

    py_version = platform.python_version()
    system_name = platform.system()
    date_time = strftime('%m/%d/%Y %I:%M:%S %p')
    logger.info("\n"
                rf"    _             | Version {version}" + '\n'
                rf"   /_\  _ __  ___ | Python {py_version} on {system_name}, {date_time}" + '\n'
                r"  / _ \| '  \(_-< | " + "\n"
                r' /_/ \_\_|_|_/__/ | This program comes with ABSOLUTELY NO WARRANTY.' + '\n')


def main():
    """
    Entry point of the AMS command-line interface.
    """

    parser = create_parser()
    args = parser.parse_args()

    # Set up logging
    config_logger(stream=True,
                  stream_level=args.verbose,
                  file=True,
                  log_path=get_log_dir(),
                  )
    logger.debug(args)

    module = importlib.import_module('ams.main')

    if args.command in ('doc', 'misc'):
        pass
    elif args.command == 'run' and args.no_preamble is True:
        pass
    else:
        preamble()

    # Run the command
    if args.command is None:
        parser.parse_args(sys.argv.append('--help'))

    else:
        cmd = args.command
        for fullcmd, aliases in command_aliases.items():
            if cmd in aliases:
                cmd = fullcmd

        func = getattr(module, cmd)
        return func(cli=True, **vars(args))



"""
Module for system matrix make.
"""

import logging
import os
import sys
from typing import Optional

import numpy as np

from andes.thirdparty.npfunc import safe_div
from andes.shared import tqdm, tqdm_nb
from andes.utils.misc import elapsed, is_notebook

from ams.opt import Param
from ams.shared import pd, sps

logger = logging.getLogger(__name__)


class MParam(Param):
    """
    Class for matrix parameters built from the system.

    MParam is designed to be a subclass of RParam for routine parameters
    management.

    Parameters
    ----------
    name : str, optional
        Name of this parameter. If not provided, `name` will be set
        to the attribute name.
    tex_name : str, optional
        LaTeX-formatted parameter name. If not provided, `tex_name`
        will be assigned the same as `name`.
    info : str, optional
        A description of this parameter
    unit : str, optional
        Unit of the parameter.
    v : np.ndarray, optional
        Matrix value of the parameter.
    owner : object, optional
        Owner of the MParam, usually the MatProcessor instance.
    sparse : bool, optional
        If True, the matrix is stored in sparse format.
    col_names : list, optional
        Column names of the matrix.
    row_names : list, optional
        Row names of the matrix.
    """

    def __init__(self,
                 name: Optional[str] = None,
                 tex_name: Optional[str] = None,
                 info: Optional[str] = None,
                 unit: Optional[str] = None,
                 v: Optional[np.ndarray] = None,
                 owner: Optional[object] = None,
                 sparse: Optional[bool] = False,
                 col_names: Optional[list] = None,
                 row_names: Optional[list] = None,
                 ):
        self.name = name
        self.tex_name = tex_name if (tex_name is not None) else name
        self.info = info
        self.unit = unit
        self._v = v
        self.sparse = sparse
        self.owner = owner
        self.col_names = col_names
        self.row_names = row_names

    def export_csv(self, path=None):
        """
        Export the matrix to a CSV file.

        Parameters
        ----------
        path : str, optional
            Path to the output CSV file.

        Returns
        -------
        str
            The path of the exported csv file
        """

        if path is None:
            if self.owner.system.files.fullname is None:
                logger.info("Input file name not detacted. Using `Untitled`.")
                file_name = f'Untitled_{self.name}'
            else:
                file_name = os.path.splitext(self.owner.system.files.fullname)[0]
                file_name += f'_{self.name}'
            path = os.path.join(os.getcwd(), file_name + '.csv')
        else:
            file_name = os.path.splitext(os.path.basename(path))[0]

        pd.DataFrame(data=self.v, columns=self.col_names, index=self.row_names).to_csv(path)

        return file_name + '.csv'

    @property
    def v(self):
        """
        Return the value of the parameter.
        """
        # NOTE: scipy.sparse matrix will return 2D array
        # so we squeeze it here if only one row
        if isinstance(self._v, (sps.csr_matrix, sps.lil_matrix, sps.csc_matrix)):
            out = self._v.toarray()
            if out.shape[0] == 1:
                return np.squeeze(out)
            else:
                return out
        return self._v

    @property
    def shape(self):
        """
        Return the shape of the parameter.
        """
        return self.v.shape

    @property
    def n(self):
        """
        Return the szie of the parameter.
        """
        return len(self.v)

    @property
    def class_name(self):
        """
        Return the class name
        """
        return self.__class__.__name__


class MatProcessor:
    """
    Class for matrices processing in AMS system.
    The connectivity matrices `Cft`, `Cg`, `Cl`, and `Csh` ***have taken*** the
    devices connectivity into account.

    The MParams' row names and col names are assigned in `System.setup()`.
    """

    def __init__(self, system):
        self.system = system
        self.initialized = False
        self.pbar = None

        self.Cft = MParam(name='Cft', tex_name=r'C_{ft}',
                          info='Line connectivity matrix',
                          v=None, sparse=True, owner=self)
        self.CftT = MParam(name='CftT', tex_name=r'C_{ft}^{T}',
                           info='Line connectivity matrix transpose',
                           v=None, sparse=True, owner=self)
        self.Cg = MParam(name='Cg', tex_name=r'C_g',
                         info='Generator connectivity matrix',
                         v=None, sparse=True, owner=self)
        self.Cl = MParam(name='Cl', tex_name=r'Cl',
                         info='Load connectivity matrix',
                         v=None, sparse=True, owner=self)
        self.Csh = MParam(name='Csh', tex_name=r'C_{sh}',
                          info='Shunt connectivity matrix',
                          v=None, sparse=True, owner=self)

        self.Bbus = MParam(name='Bbus', tex_name=r'B_{bus}',
                           info='Bus admittance matrix',
                           v=None, sparse=True, owner=self)
        self.Bf = MParam(name='Bf', tex_name=r'B_{f}',
                         info='Bf matrix',
                         v=None, sparse=True, owner=self)
        self.Pbusinj = MParam(name='Pbusinj', tex_name=r'P_{bus}^{inj}',
                              info='Bus power injection vector',
                              v=None, sparse=False, owner=self)
        self.Pfinj = MParam(name='Pfinj', tex_name=r'P_{f}^{inj}',
                            info='Line power injection vector',
                            v=None, sparse=False, owner=self)

        self.PTDF = MParam(name='PTDF', tex_name=r'P_{TDF}',
                           info='Power transfer distribution factor',
                           v=None, sparse=False, owner=self)
        self.LODF = MParam(name='LODF', tex_name=r'O_{TDF}',
                           info='Line outage distribution factor',
                           v=None, sparse=False, owner=self)

    def build(self, force=False):
        """
        Build the system matrices.
        It build connectivity matrices first: Cg, Cl, Csh, Cft, and CftT.
        Then build bus matrices: Bf, Bbus, Pfinj, and Pbusinj.

        Parameters
        ----------
        force : bool, optional
            If True, force to rebuild the matrices. Default is False.

        Notes
        -----
        Generator online status is NOT considered in its connectivity matrix.
        The same applies for load, line, and shunt.

        Returns
        -------
        initialized : bool
            True if the matrices are built successfully.
        """
        if not force and self.initialized:
            logger.debug("System matrices are already built.")
            return self.initialized

        t_mat, _ = elapsed()
        logger.warning("Building system matrices")
        # --- connectivity matrices ---
        _ = self.build_cg()
        _ = self.build_cl()
        _ = self.build_csh()
        _ = self.build_cft()

        # --- bus matrices ---
        _ = self.build_bf()
        _ = self.build_bbus()
        _ = self.build_pfinj()
        _ = self.build_pbusinj()
        _, s_mat = elapsed(t_mat)

        logger.debug(f" -> System matrices built in {s_mat}")
        self.initialized = True
        return self.initialized

    @property
    def class_name(self):
        """
        Return the class name
        """
        return self.__class__.__name__

    @property
    def n(self):
        """
        To fit the RParam style.
        """
        return 2

    def build_cg(self):
        """
        Build generator connectivity matrix Cg, and store it in the MParam `Cg`.

        Returns
        -------
        Cg : scipy.sparse.csr_matrix
            Generator connectivity matrix.
        """
        system = self.system

        # common variables
        nb = system.Bus.n
        ng = system.StaticGen.n

        # bus indices: idx -> uid
        idx_gen = system.StaticGen.get_all_idxes()
        u_gen = system.StaticGen.get(src='u', attr='v', idx=idx_gen)
        on_gen = np.flatnonzero(u_gen)  # uid of online generators
        on_gen_idx = [idx_gen[i] for i in on_gen]  # idx of online generators
        on_gen_bus = system.StaticGen.get(src='bus', attr='v', idx=on_gen_idx)

        row = np.array([system.Bus.idx2uid(x) for x in on_gen_bus])
        col = np.array([idx_gen.index(x) for x in on_gen_idx])
        self.Cg._v = sps.csr_matrix((np.ones(len(on_gen_idx)), (row, col)), (nb, ng))
        return self.Cg._v

    def build_cl(self):
        """
        Build load connectivity matrix Cl, and store it in the MParam `Cl`.

        Returns
        -------
        Cl : scipy.sparse.csr_matrix
            Load connectivity matrix.
        """
        system = self.system

        # common variables
        nb = system.Bus.n
        npq = system.PQ.n

        # load indices: idx -> uid
        idx_load = system.PQ.idx.v
        u_load = system.PQ.get(src='u', attr='v', idx=idx_load)
        on_load = np.flatnonzero(u_load)
        on_load_idx = [idx_load[i] for i in on_load]
        on_load_bus = system.PQ.get(src='bus', attr='v', idx=on_load_idx)

        row = np.array([system.Bus.idx2uid(x) for x in on_load_bus])
        col = np.array([system.PQ.idx2uid(x) for x in on_load_idx])
        self.Cl._v = sps.csr_matrix((np.ones(len(on_load_idx)), (row, col)), (nb, npq))
        return self.Cl._v

    def build_csh(self):
        """
        Build shunt connectivity matrix Csh, and store it in the MParam `Csh`.

        Returns
        -------
        Csh : spmatrix
            Shunt connectivity matrix.
        """
        system = self.system

        # common variables
        nb = system.Bus.n
        nsh = system.Shunt.n

        # shunt indices: idx -> uid
        idx_shunt = system.Shunt.idx.v
        u_shunt = system.Shunt.get(src='u', attr='v', idx=idx_shunt)
        on_shunt = np.flatnonzero(u_shunt)
        on_shunt_idx = [idx_shunt[i] for i in on_shunt]
        on_shunt_bus = system.Shunt.get(src='bus', attr='v', idx=on_shunt_idx)

        row = np.array([system.Bus.idx2uid(x) for x in on_shunt_bus])
        col = np.array([system.Shunt.idx2uid(x) for x in on_shunt_idx])
        self.Csh._v = sps.csr_matrix((np.ones(len(on_shunt_idx)), (row, col)), (nb, nsh))
        return self.Csh._v

    def build_cft(self):
        """
        Build line connectivity matrix Cft and its transpose CftT.
        The Cft and CftT are stored in the MParam `Cft` and `CftT`, respectively.

        Returns
        -------
        Cft : scipy.sparse.csr_matrix
            Line connectivity matrix.
        """
        system = self.system

        # common variables
        nb = system.Bus.n
        nl = system.Line.n

        # line indices: idx -> uid
        idx_line = system.Line.idx.v
        u_line = system.Line.get(src='u', attr='v', idx=idx_line)
        on_line = np.flatnonzero(u_line)
        on_line_idx = [idx_line[i] for i in on_line]
        on_line_bus1 = system.Line.get(src='bus1', attr='v', idx=on_line_idx)
        on_line_bus2 = system.Line.get(src='bus2', attr='v', idx=on_line_idx)

        data_line = np.ones(2*len(on_line_idx))
        data_line[len(on_line_idx):] = -1
        row_line = np.array([system.Bus.idx2uid(x) for x in on_line_bus1 + on_line_bus2])
        col_line = np.array([system.Line.idx2uid(x) for x in on_line_idx + on_line_idx])
        self.Cft._v = sps.csr_matrix((data_line, (row_line, col_line)), (nb, nl))
        self.CftT._v = self.Cft._v.T
        return self.Cft._v

    def build_bf(self):
        """
        Build DC Bf matrix and store it in the MParam `Bf`.

        Returns
        -------
        Bf : scipy.sparse.csr_matrix
            Bf matrix.
        """
        system = self.system

        # common variables
        nb = system.Bus.n
        nl = system.Line.n

        # line parameters
        idx_line = system.Line.idx.v
        b = self._calc_b()

        # build Bf such that Bf * Va is the vector of real branch powers injected
        # at each branch's "from" bus
        f = system.Bus.idx2uid(system.Line.get(src='bus1', attr='v', idx=idx_line))
        t = system.Bus.idx2uid(system.Line.get(src='bus2', attr='v', idx=idx_line))
        ir = np.r_[range(nl), range(nl)]  # double set of row indices
        self.Bf._v = sps.csr_matrix((np.r_[b, -b], (ir, np.r_[f, t])), (nl, nb))
        return self.Bf._v

    def build_bbus(self):
        """
        Build Bdc matrix and store it in the MParam `Bbus`.

        Returns
        -------
        Bdc : scipy.sparse.csr_matrix
            DC bus admittance matrix.
        """
        self.Bbus._v = self.Cft._v * self.Bf._v
        return self.Bbus._v

    def build_pfinj(self):
        """
        Build DC Pfinj vector and store it in the MParam `Pfinj`.

        Returns
        -------
        Pfinj : np.ndarray
            Line power injection vector.
        """
        idx_line = self.system.Line.idx.v
        b = self._calc_b()
        phi = self.system.Line.get(src='phi', attr='v', idx=idx_line)
        self.Pfinj._v = b * (-phi)
        return self.Pfinj._v

    def build_pbusinj(self):
        """
        Build DC Pbusinj vector and store it in the MParam `Pbusinj`.

        Returns
        -------
        Pbusinj : np.ndarray
            Bus power injection vector.
        """
        self.Pbusinj._v = self.Cft._v * self.Pfinj._v
        return self.Pbusinj._v

    def _calc_b(self):
        """
        Calculate DC series susceptance for each line.

        Returns
        -------
        b : np.ndarray
            Series susceptance for each line.
        """
        system = self.system
        nl = system.Line.n

        # line parameters
        idx_line = system.Line.idx.v
        x = system.Line.get(src='x', attr='v', idx=idx_line)
        u_line = system.Line.get(src='u', attr='v', idx=idx_line)
        b = u_line / x  # series susceptance

        # in DC, tap is assumed to be 1
        tap0 = system.Line.get(src='tap', attr='v', idx=idx_line)
        tap = np.ones(nl)
        i = np.flatnonzero(tap0)
        tap[i] = tap0[i]  # assign non-zero tap ratios
        b = b / tap  # adjusted series susceptance

        return b

    def build_ptdf(self, line=None, no_store=False,
                   incremental=False, step=1000, no_tqdm=False,
                   permc_spec=None, use_umfpack=True):
        """
        Build the Power Transfer Distribution Factor (PTDF) matrix and optionally store it in `MParam.PTDF`.

        PTDF[m, n] represents the increased line flow on line `m` for a 1 p.u. power injection at bus `n`.
        It is similar to the Generation Shift Factor (GSF).

        Note: There may be minor discrepancies between PTDF-based line flow and DCOPF-calculated line flow.

        For large cases, use `incremental=True` to calculate the sparse PTDF in chunks, which will be stored
        as a `scipy.sparse.lil_matrix`. In this mode, the PTDF is calculated in chunks, and a progress bar
        will be shown unless `no_tqdm=True`.

        Parameters
        ----------
        line : int, str, list, optional
            Line indices for which the PTDF is calculated. If specified, the PTDF will not be stored in `MParam`.
        no_store : bool, optional
            If False, store the PTDF in `MatProcessor.PTDF._v`. Default is False.
        incremental : bool, optional
            If True, calculate the sparse PTDF in chunks to save memory. Default is False.
        step : int, optional
            Step size for incremental calculation. Default is 1000.
        no_tqdm : bool, optional
            If True, disable the progress bar. Default is False.
        permc_spec : str, optional
            Column permutation strategy for sparsity preservation. Default is 'COLAMD'.
        use_umfpack : bool, optional
            If True, use UMFPACK as the solver. Effective only when (`incremental=True`)
            & (`line` contains a single line or `step` is 1). Default is True.

        Returns
        -------
        PTDF : np.ndarray or scipy.sparse.lil_matrix
            Power transfer distribution factor.

        References
        ----------
        [1] PowerWorld Documentation, Power Transfer Distribution Factors, [Online]

        Available:

        https://www.powerworld.com/WebHelp/Content/MainDocumentation_HTML/Power_Transfer_Distribution_Factors.htm
        """
        system = self.system

        # use first slack bus as reference slack bus
        slack = system.Slack.bus.v[0]
        noslack = [system.Bus.idx2uid(bus) for bus in system.Bus.idx.v if bus != slack]

        # use first bus for voltage angle reference
        noref_idx = system.Bus.idx.v[1:]
        noref = system.Bus.idx2uid(noref_idx)

        if line is None:
            luid = system.Line.idx2uid(system.Line.idx.v)
        elif isinstance(line, (int, str)):
            try:
                luid = [system.Line.idx2uid(line)]
            except ValueError:
                raise ValueError(f"Line {line} not found.")
        elif isinstance(line, list):
            luid = system.Line.idx2uid(line)

        # build other matrices if not built
        if not self.initialized:
            logger.debug("System matrices are not built. Building now.")
            self.build()

        nbus = system.Bus.n
        nline = len(luid)

        Bbus = self.Bbus._v
        Bf = self.Bf._v

        if incremental:
            # initialize progress bar
            if is_notebook():
                self.pbar = tqdm_nb(total=100, unit='%', file=sys.stdout,
                                    disable=no_tqdm)
            else:
                self.pbar = tqdm(total=100, unit='%', ncols=80, ascii=True,
                                 file=sys.stdout, disable=no_tqdm)

            self.pbar.update(0)
            last_pc = 0

            H = sps.lil_matrix((nline, system.Bus.n))

            # NOTE: for PTDF, we are building rows by rows
            for start in range(0, nline, step):
                end = min(start + step, nline)
                sol = sps.linalg.spsolve(A=Bbus[np.ix_(noslack, noref)].T,
                                         b=Bf[np.ix_(luid[start:end], noref)].T,
                                         permc_spec=permc_spec,
                                         use_umfpack=use_umfpack).T
                H[start:end, noslack] = sol

                # show progress in percentage
                perc = np.round(min((end / nline) * 100, 100), 2)

                perc_diff = perc - last_pc
                if perc_diff >= 1:
                    self.pbar.update(perc_diff)
                    last_pc = perc

            # finish progress bar
            self.pbar.update(100 - last_pc)
            # removed `pbar` so that System object can be serialized
            self.pbar.close()
            self.pbar = None
        else:
            H = np.zeros((nline, nbus))
            H[:, noslack] = np.linalg.solve(Bbus.todense()[np.ix_(noslack, noref)].T,
                                            Bf.todense()[np.ix_(luid, noref)].T).T

        # reshape results into 1D array if only one line
        if isinstance(line, (int, str)):
            H = H[0, :]

        if (not no_store) & (line is None):
            self.PTDF._v = H

        return H

    def build_lodf(self, line=None, no_store=False,
                   incremental=False, step=1000, no_tqdm=False):
        """
        Build the Line Outage Distribution Factor matrix and store it in the
        MParam `LODF`.

        `LODF[m, n]` means the increased line flow on line `m` when there is
        1 p.u. line flow decrease on line `n` due to line `n` outage.
        It is also referred to as Branch Outage Distribution Factor (BODF).

        It requires DC PTDF and Cft.

        For large cases where memory is a concern, use `incremental=True` to
        calculate the sparse LODF in chunks in the format of scipy.sparse.lil_matrix.

        Parameters
        ----------
        line: int, str, list, optional
            Lines index for which the LODF is calculated. It takes both single
            or multiple line indices. Note that if `line` is given, the LODF will
            not be stored in the MParam.
        no_store : bool, optional
            If False, the LODF will be stored into `MatProcessor.LODF._v`.
        incremental : bool, optional
            If True, the sparse LODF will be calculated in chunks to save memory.
        step : int, optional
            Step for incremental calculation.
        no_tqdm : bool, optional
            If True, the progress bar will be disabled.

        Returns
        -------
        LODF : np.ndarray, scipy.sparse.lil_matrix
            Line outage distribution factor.

        References
        ----------
        [1] PowerWorld Documentation, Line Outage Distribution Factors, [Online]

        Available:

        https://www.powerworld.com/WebHelp/Content/MainDocumentation_HTML/Line_Outage_Distribution_Factors_LODFs.htm
        """
        system = self.system

        if line is None:
            luid = system.Line.idx2uid(system.Line.idx.v)
        elif isinstance(line, (int, str)):
            try:
                luid = [system.Line.idx2uid(line)]
            except ValueError:
                raise ValueError(f"Line {line} not found.")
        elif isinstance(line, list):
            luid = system.Line.idx2uid(line)

        # NOTE: here we use nbranch to differentiate it with nline
        nbranch = system.Line.n
        nline = len(luid)

        ptdf = self.PTDF._v
        # build PTDF if not built
        if self.PTDF._v is None:
            ptdf = self.build_ptdf(no_store=True, incremental=incremental, step=step)
        if incremental and isinstance(self.PTDF._v, np.ndarray):
            ptdf = sps.lil_matrix(self.PTDF._v)

        if incremental | (isinstance(ptdf, sps.spmatrix)):
            # initialize progress bar
            if is_notebook():
                self.pbar = tqdm_nb(total=100, unit='%', file=sys.stdout,
                                    disable=no_tqdm)
            else:
                self.pbar = tqdm(total=100, unit='%', ncols=80, ascii=True,
                                 file=sys.stdout, disable=no_tqdm)

            self.pbar.update(0)
            last_pc = 0

            LODF = sps.lil_matrix((nbranch, nline))

            # NOTE: for LODF, we are doing it columns by columns
            # reshape luid to list of list by step
            luidp = [luid[i:i + step] for i in range(0, len(luid), step)]
            for luidi in luidp:
                H_chunk = ptdf @ self.Cft._v[:, luidi]
                h_chunk = H_chunk.diagonal(-luidi[0])
                rden = safe_div(np.ones(H_chunk.shape),
                                np.tile(np.ones_like(h_chunk) - h_chunk, (nbranch, 1)))
                H_chunk = H_chunk.multiply(rden).tolil()
                # NOTE: use lil_matrix to set diagonal values as -1
                rsid = sps.diags(H_chunk.diagonal(-luidi[0])) + sps.eye(H_chunk.shape[1])
                if H_chunk.shape[0] > rsid.shape[0]:
                    Rsid = sps.lil_matrix(H_chunk.shape)
                    Rsid[luidi, :] = rsid
                else:
                    Rsid = rsid
                H_chunk = H_chunk - Rsid
                LODF[:, [luid.index(i) for i in luidi]] = H_chunk

                # show progress in percentage
                perc = np.round(min((luid.index(luidi[-1]) / nline) * 100, 100), 2)

                perc_diff = perc - last_pc
                if perc_diff >= 1:
                    self.pbar.update(perc_diff)
                    last_pc = perc

            # finish progress bar
            self.pbar.update(100 - last_pc)
            # removed `pbar` so that System object can be serialized
            self.pbar.close()
            self.pbar = None
        else:
            H = ptdf @ self.Cft._v[:, luid]
            h = np.diag(H, -luid[0])
            LODF = safe_div(H,
                            np.tile(np.ones_like(h) - h, (nbranch, 1)))
            # # NOTE: reset the diagonal elements to -1
            rsid = np.diag(np.diag(LODF, -luid[0])) + np.eye(nline, nline)
            if LODF.shape[0] > rsid.shape[0]:
                Rsid = np.zeros_like(LODF)
                Rsid[luid, :] = rsid
            else:
                Rsid = rsid
            LODF = LODF - Rsid

        # reshape results into 1D array if only one line
        if isinstance(line, (int, str)):
            LODF = LODF[:, 0]

        if (not no_store) & (line is None):
            self.LODF._v = LODF
        return LODF

    def build_otdf(self, line=None):
        """
        Build the Outrage Transfer Distribution Factor (OTDF) matrix for line
        k outage: :math:`OTDF_k = PTDF + LODF[:, k] @ PTDF[k, ]`.

        OTDF_k[m, n] means the increased line flow on line `m` when there is
        1 p.u. power injection at bus `n` when line `k` is outage.

        Note that the OTDF is not stored in the MatProcessor.

        Parameters
        ----------
        line : int, str, list, optional
            Lines index for which the OTDF is calculated. It takes both single
            or multiple line indices.
            If not given, the first line is used by default.

        Returns
        -------
        OTDF : np.ndarray, scipy.sparse.csr_matrix
            Line outage distribution factor.

        References
        ----------
        [1] PowerWorld Documentation, Line Outage Distribution Factors, [Online]

        Available:

        https://www.powerworld.com/WebHelp/Content/MainDocumentation_HTML/Line_Outage_Distribution_Factors_LODFs.htm
        """
        if (self.PTDF._v is None) or (self.LODF._v is None):
            raise ValueError("Internal PTDF and LODF are not available. Please build them first.")

        ptdf = self.PTDF._v
        lodf = self.LODF._v

        if line is None:
            luid = [0]
        elif isinstance(line, (int, str)):
            try:
                luid = [self.system.Line.idx2uid(line)]
            except ValueError:
                raise ValueError(f"Line {line} not found.")
        elif isinstance(line, list):
            luid = self.system.Line.idx2uid(line)

        otdf = ptdf + lodf[:, luid] @ ptdf[luid, :]
        return otdf



"""
Documenter class for AMS models.
"""
import inspect
import re

import logging
from collections import OrderedDict
from andes.core.documenter import Documenter as andes_Documenter
from andes.utils.tab import make_doc_table, math_wrap

logger = logging.getLogger(__name__)


def disable_method(func):
    def wrapper(*args, **kwargs):
        msg = f"Method `{func.__name__}` is included in ANDES Documenter but not supported in AMS Documenter."
        logger.warning(msg)
        return None
    return wrapper


def disable_methods(methods):
    for method in methods:
        setattr(Documenter, method, disable_method(getattr(Documenter, method)))


class Documenter(andes_Documenter):
    """
    Helper class for documenting models.

    Parameters
    ----------
    parent : Model
        The `Model` instance to document
    """

    def __init__(self, parent):
        self.parent = parent
        self.system = parent.system
        self.class_name = parent.class_name
        self.config = parent.config
        self.params = parent.params
        self.algebs = parent.algebs
        self.services = parent.services

        func_to_disable = ['_init_doc', '_eq_doc',
                           '_discrete_doc', '_block_doc']
        disable_methods(func_to_disable)

    def get(self, max_width=78, export='plain'):
        """
        Return the model documentation in table-formatted string.

        Parameters
        ----------
        max_width : int
            Maximum table width. Automatically et to 0 if format is ``rest``.
        export : str, ('plain', 'rest')
            Export format. Use fancy table if is ``rest``.

        Returns
        -------
        str
            A string with the documentations.
        """
        out = ''
        if export == 'rest':
            max_width = 0
            model_header = '-' * 80 + '\n'
            out += f'.. _{self.class_name}:\n\n'
        else:
            model_header = ''

        if export == 'rest':
            out += model_header + f'{self.class_name}\n' + model_header
        else:
            out += model_header + f'Model <{self.class_name}> in Group <{self.parent.group}>\n' + model_header

        if self.parent.__doc__ is not None:
            out += inspect.cleandoc(self.parent.__doc__)
        out += '\n\n'  # this fixes the indentation for the next line

        # add tables
        out += self._param_doc(max_width=max_width, export=export)
        out += self._var_doc(max_width=max_width, export=export)
        out += self._service_doc(max_width=max_width, export=export)
        # TODO: fix and add the config doc later on
        # out += self.config.doc(max_width=max_width, export=export)

        return out

    def _var_doc(self, max_width=78, export='plain'):
        # variable documentation
        if len(self.algebs) == 0:
            return ''

        names, symbols, units = list(), list(), list()
        info = list()
        units_rest, ty = list(), list()

        for p in self.algebs.values():
            names.append(p.name)
            ty.append(p.class_name)
            info.append(p.info if p.info else '')
            units.append(p.unit if p.unit else '')
            units_rest.append(f'*{p.unit}*' if p.unit else '')

        title = 'Variables'

        # replace with latex math expressions if export is ``rest``
        if export == 'rest':
            symbols = [item.tex_name for item in self.algebs.values()]
            symbols = math_wrap(symbols, export=export)
            title = 'Variables\n---------'

        plain_dict = OrderedDict([('Name', names),
                                  ('Type', ty),
                                  ('Description', info),
                                  ('Unit', units)])

        rest_dict = OrderedDict([('Name', names),
                                 ('Symbol', symbols),
                                 ('Type', ty),
                                 ('Description', info),
                                 ('Unit', units_rest)])

        return make_doc_table(title=title,
                              max_width=max_width,
                              export=export,
                              plain_dict=plain_dict,
                              rest_dict=rest_dict)

    def _service_doc(self, max_width=78, export='plain'):
        if len(self.services) == 0:
            return ''

        names, symbols = list(), list()
        info = list()
        class_names = list()

        for p in self.services.values():
            names.append(p.name)
            class_names.append(p.class_name)
            info.append(p.info if p.info else '')
            symbols.append(p.tex_name if p.tex_name is not None else '')

        title = 'Services'
        if export == 'rest':
            symbols = math_wrap(symbols, export=export)
            title = 'Services\n----------'

        plain_dict = OrderedDict([('Name', names),
                                  ('Description', info),
                                  ('Type', class_names)])

        rest_dict = OrderedDict([('Name', names),
                                 ('Description', info),
                                 ('Symbol', symbols),
                                 ('Type', class_names)])

        return make_doc_table(title=title,
                              max_width=max_width,
                              export=export,
                              plain_dict=plain_dict,
                              rest_dict=rest_dict)


class RDocumenter:
    """
    Helper class for documenting routines.

    Parameters
    ----------
    parent : Model
        The `Model` instance to document
    """

    def __init__(self, parent):
        self.parent = parent
        self.system = parent.system
        self.class_name = parent.class_name
        self.config = parent.config
        self.services = parent.services
        self.rparams = parent.rparams
        self.vars = parent.vars
        self.constrs = parent.constrs
        self.obj = parent.obj

    def get(self, max_width=78, export='plain'):
        """
        Return the routine documentation in table-formatted string.

        Parameters
        ----------
        max_width : int
            Maximum table width. Automatically et to 0 if format is ``rest``.
        export : str, ('plain', 'rest')
            Export format. Use fancy table if is ``rest``.

        Returns
        -------
        str
            A string with the documentations.
        """
        out = ''
        if export == 'rest':
            max_width = 0
            model_header = '-' * 80 + '\n'
            out += f'.. _{self.class_name}:\n\n'
        else:
            model_header = ''

        if export == 'rest':
            out += model_header + f'{self.class_name}\n' + model_header
        else:
            out += model_header + f'Routine <{self.class_name}> in Type <{self.parent.type}>\n' + model_header

        if self.parent.__doc__ is not None:
            out += inspect.cleandoc(self.parent.__doc__)
        out += '\n\n'  # this fixes the indentation for the next line

        # add tables
        self.parent.syms.generate_symbols()
        out += self._obj_doc(max_width=max_width, export=export)
        out += self._expr_doc(max_width=max_width, export=export)
        out += self._constr_doc(max_width=max_width, export=export)
        out += self._var_doc(max_width=max_width, export=export)
        out += self._exprc_doc(max_width=max_width, export=export)
        out += self._service_doc(max_width=max_width, export=export)
        out += self._param_doc(max_width=max_width, export=export)
        out += self.config.doc(max_width=max_width, export=export)

        return out

    def _service_doc(self, max_width=78, export='plain'):
        # routine service documentation
        if len(self.services) == 0:
            return ''

        names, symbols = list(), list()
        info = list()
        class_names = list()

        for p in self.services.values():
            names.append(p.name)
            info.append(p.info if p.info else '')
            class_names.append(p.class_name)

        title = 'Services'

        # replace with latex math expressions if export is ``rest``
        if export == 'rest':
            symbols = [item.tex_name for item in self.services.values()]
            symbols = math_wrap(symbols, export=export)
            title = 'Services\n---------'

        plain_dict = OrderedDict([('Name', names),
                                  ('Description', info),
                                  ('Type', class_names)])

        rest_dict = OrderedDict([('Name', names),
                                 ('Symbol', symbols),
                                 ('Description', info),
                                 ('Type', class_names)])

        return make_doc_table(title=title,
                              max_width=max_width,
                              export=export,
                              plain_dict=plain_dict,
                              rest_dict=rest_dict)

    def _constr_doc(self, max_width=78, export='plain'):
        # constraint documentation
        if len(self.constrs) == 0:
            return ''

        # prepare temporary lists
        names, class_names, info = list(), list(), list()

        for p in self.constrs.values():
            names.append(p.name)
            class_names.append(p.class_name)
            info.append(p.info if p.info else '')

        # expressions based on output format
        expressions = []
        if export == 'rest':
            for p in self.constrs.values():
                expr = _tex_pre(self, p, self.parent.syms.tex_map)
                if p.is_eq:
                    expr = f'{expr} = 0'
                else:
                    expr = f'{expr} \\leq 0'
                logger.debug(f'{p.name} math: {expr}')
                expressions.append(expr)

            title = 'Constraints\n----------------------------------'
        else:
            title = 'Constraints'
        expressions = math_wrap(expressions, export=export)

        plain_dict = OrderedDict([('Name', names),
                                  ('Description', info),
                                  ])

        rest_dict = OrderedDict([('Name', names),
                                 ('Description', info),
                                 ('Expression', expressions),
                                 ])

        # convert to rows and export as table
        return make_doc_table(title=title,
                              max_width=max_width,
                              export=export,
                              plain_dict=plain_dict,
                              rest_dict=rest_dict)

    def _expr_doc(self, max_width=78, export='plain'):
        # Expression documentation
        if len(self.parent.exprs) == 0:
            return ''

        # prepare temporary lists
        names, info = list(), list()
        units, sources, units_rest = list(), list(), list()

        for p in self.parent.exprs.values():
            names.append(p.name)
            info.append(p.info if p.info else '')
            units.append(p.unit if p.unit else '')
            units_rest.append(f'*{p.unit}*' if p.unit else '')

            slist = []
            if p.owner is not None and p.src is not None:
                slist.append(f'{p.owner.class_name}.{p.src}')
            elif p.owner is not None and p.src is None:
                slist.append(f'{p.owner.class_name}')
            sources.append(','.join(slist))

        # expressions based on output format
        expressions = []
        if export == 'rest':
            for p in self.parent.exprs.values():
                expr = _tex_pre(self, p, self.parent.syms.tex_map)
                logger.debug(f'{p.name} math: {expr}')
                expressions.append(expr)

            title = 'Expressions\n----------------------------------'
        else:
            title = 'Expressions'
        expressions = math_wrap(expressions, export=export)

        plain_dict = OrderedDict([('Name', names),
                                  ('Description', info),
                                  ('Unit', units),
                                  ])
        rest_dict = OrderedDict([('Name', names),
                                 ('Description', info),
                                 ('Expression', expressions),
                                 ('Unit', units_rest),
                                 ('Source', sources),
                                 ])

        # convert to rows and export as table
        return make_doc_table(title=title,
                              max_width=max_width,
                              export=export,
                              plain_dict=plain_dict,
                              rest_dict=rest_dict)

    def _exprc_doc(self, max_width=78, export='plain'):
        # ExpressionCalc documentation
        if len(self.parent.exprcs) == 0:
            return ''

        # prepare temporary lists
        names, info = list(), list()
        units, sources, units_rest = list(), list(), list()

        for p in self.parent.exprcs.values():
            names.append(p.name)
            info.append(p.info if p.info else '')
            units.append(p.unit if p.unit else '')
            units_rest.append(f'*{p.unit}*' if p.unit else '')

            slist = []
            if p.owner is not None and p.src is not None:
                slist.append(f'{p.owner.class_name}.{p.src}')
            elif p.owner is not None and p.src is None:
                slist.append(f'{p.owner.class_name}')
            sources.append(','.join(slist))

        # expressions based on output format
        expressions = []
        if export == 'rest':
            for p in self.parent.exprcs.values():
                expr = _tex_pre(self, p, self.parent.syms.tex_map)
                logger.debug(f'{p.name} math: {expr}')
                expressions.append(expr)

            title = 'ExpressionCalcs\n----------------------------------'
        else:
            title = 'ExpressionCalcs'
        expressions = math_wrap(expressions, export=export)

        plain_dict = OrderedDict([('Name', names),
                                  ('Description', info),
                                  ('Unit', units),
                                  ])
        rest_dict = OrderedDict([('Name', names),
                                 ('Description', info),
                                 ('Expression', expressions),
                                 ('Unit', units_rest),
                                 ('Source', sources),
                                 ])

        # convert to rows and export as table
        return make_doc_table(title=title,
                              max_width=max_width,
                              export=export,
                              plain_dict=plain_dict,
                              rest_dict=rest_dict)

    def _obj_doc(self, max_width=78, export='plain'):
        # variable documentation
        if self.parent.obj is None:
            return ''

        # prepare temporary lists
        units, units_rest = list(), list()

        p = self.parent.obj
        units.append(p.unit if p.unit else '')
        units_rest.append(f'*{p.unit}*' if p.unit else '')

        # expressions based on output format
        expr = _tex_pre(self, p, self.parent.syms.tex_map)
        expr = p.sense + '. ' + expr  # minimize or maximize
        expr = [expr]
        if export == 'rest':
            expr = math_wrap(expr, export=export)
            title = 'Objective\n----------------------------------'
        else:
            title = 'Objective'

        plain_dict = OrderedDict([('Unit', units),])

        rest_dict = OrderedDict([('Unit', units_rest),
                                 ('Expression', expr)])

        # convert to rows and export as table
        return make_doc_table(title=title,
                              max_width=max_width,
                              export=export,
                              plain_dict=plain_dict,
                              rest_dict=rest_dict)

    def _var_doc(self, max_width=78, export='plain'):
        # NOTE: this is for the optimization variables
        # not the _var_doc for ANDES parameters
        # variable documentation
        if len(self.vars) == 0:
            return ''

        # prepare temporary lists
        names, units, class_names = list(), list(), list()
        properties, info = list(), list()
        sources = list()
        units_rest = list()

        for p in self.vars.values():
            names.append(p.name)
            class_names.append(p.class_name)
            info.append(p.info if p.info else '')
            # defaults.append(p.default if p.default is not None else '')
            units.append(f'{p.unit}' if p.unit else '')
            units_rest.append(f'*{p.unit}*' if p.unit else '')

            # collect properties defined in config
            plist = []
            for key, val in p.config.as_dict().items():
                if val is True:
                    plist.append(key)
            properties.append(','.join(plist))

            slist = []
            if p.owner is not None and p.src is not None:
                slist.append(f'{p.owner.class_name}.{p.src}')
            elif p.owner is not None and p.src is None:
                slist.append(f'{p.owner.class_name}')
            sources.append(','.join(slist))

        # symbols based on output format
        if export == 'rest':
            symbols = [item.tex_name for item in self.vars.values()]
            symbols = math_wrap(symbols, export=export)
            title = 'Vars\n----------------------------------'
        else:
            symbols = [item.name for item in self.vars.values()]
            title = 'Vars'

        plain_dict = OrderedDict([('Name', names),
                                  ('Description', info),
                                  ('Unit', units),
                                  ('Properties', properties)])

        rest_dict = OrderedDict([('Name', names),
                                 ('Symbol', symbols),
                                 ('Description', info),
                                 ('Unit', units_rest),
                                 ('Source', sources),
                                 ('Properties', properties)])

        # convert to rows and export as table
        return make_doc_table(title=title,
                              max_width=max_width,
                              export=export,
                              plain_dict=plain_dict,
                              rest_dict=rest_dict)

    def _param_doc(self, max_width=78, export='plain'):
        """
        Export formatted routine parameter documentation as a string.

        Parameters
        ----------
        max_width : int, optional = 80
            Maximum table width. If export format is ``rest`` it will be unlimited.

        export : str, optional = 'plain'
            Export format, 'plain' for plain text, 'rest' for restructuredText.

        Returns
        -------
        str
            Tabulated output in a string
        """
        if len(self.rparams) == 0:
            return ''

        # prepare temporary lists
        names, units, class_names = list(), list(), list()
        info = list()
        sources = list()
        units_rest = list()

        for p in self.rparams.values():
            names.append(p.name)
            class_names.append(p.class_name)
            info.append(p.info if p.info else '')
            # defaults.append(p.default if p.default is not None else '')
            units.append(f'{p.unit}' if p.unit else '')
            units_rest.append(f'*{p.unit}*' if p.unit else '')

            slist = []
            if p.owner is not None and p.src is not None:
                slist.append(f'{p.owner.class_name}.{p.src}')
            elif p.owner is not None and p.src is None:
                slist.append(f'{p.owner.class_name}')
            sources.append(','.join(slist))

        # symbols based on output format
        if export == 'rest':
            symbols = [item.tex_name for item in self.rparams.values()]
            symbols = math_wrap(symbols, export=export)
            title = 'Parameters\n----------------------------------'
        else:
            symbols = [item.name for item in self.rparams.values()]
            title = 'Parameters'

        plain_dict = OrderedDict([('Name', names),
                                  ('Description', info),
                                  ('Unit', units),
                                  ])

        rest_dict = OrderedDict([('Name', names),
                                 ('Symbol', symbols),
                                 ('Description', info),
                                 ('Unit', units_rest),
                                 ('Source', sources),
                                 ])

        # convert to rows and export as table
        return make_doc_table(title=title,
                              max_width=max_width,
                              export=export,
                              plain_dict=plain_dict,
                              rest_dict=rest_dict)


def _tex_pre(docm, p, tex_map):
    """
    Prepare the expression for pretty printing.

    Parameters
    ----------
    docm : Documenter
        The Documenter instance.
    p : obj or const
        The objective or constraint instance.
    tex_map : dict
        The tex map to use.
    """

    # NOTE: in the future, there might occur special math symbols
    map_before = OrderedDict([
        ('sum', 'SUM'),
        (r'\sum', 'SUM'),
        (r'\eta', 'ETA'),
        (r'\gamma', 'GAMMA'),
        (r'\theta', 'THETA'),
        (r'\phi', 'PHI'),
        (r'\frac', 'FRAC'),
        (r'\overline', 'OVERLINE'),
        (r'\underline', 'UNDERLINE'),
    ])
    map_post = OrderedDict([
        ('SUM', r'\sum'),
        ('THETA', r'\theta'),
        ('ETA', r'\eta'),
        ('GAMMA', r'\gamma'),
        ('PHI', r'\phi'),
        ('FRAC', r'\frac'),
        ('OVERLINE', r'\overline'),
        ('UNDERLINE', r'\underline'),
    ])

    expr = p.e_str

    for pattern, replacement in tex_map.items():
        for key, val in map_before.items():
            if key in replacement:
                replacement = replacement.replace(key, val)
        if r'\p' in replacement:
            continue
        try:
            expr = re.sub(pattern, replacement, expr)
        except re.error:
            expr_pattern = pattern.removeprefix('\\b').removesuffix('\\b')
            msg = f'Failed to parse <{expr_pattern}> in {docm.parent.class_name} <{p.name}>, check its tex_name.'
            logger.error(msg)
            expr = ''
        try:
            expr = expr.replace('*', ' ')
        except re.error:
            logger.error('Remains '*' in the expression.')

    for pattern, replacement in map_post.items():
        expr = expr.replace(pattern, replacement)

    return expr



"""
Python Interior Point Solver (PIPS).
"""
import logging

import numpy as np
from numpy import flatnonzero as find

import scipy.sparse as sp
from scipy.sparse import csr_matrix as c_sparse

from andes.shared import deg2rad, rad2deg

from ams.shared import inf

from ams.pypower.eps import EPS
from ams.pypower.idx import IDX
from ams.pypower.utils import sub2ind
from ams.pypower.make import makeYbus
from ams.pypower.routines.opffcns import (opf_costfcn, opf_consfcn,
                                          opf_hessfcn,)


logger = logging.getLogger(__name__)


def pips(f_fcn, x0=None, A=None, l=None, u=None, xmin=None, xmax=None,
         gh_fcn=None, hess_fcn=None, opt=None):
    """
    Primal-dual interior point method for NLP (nonlinear programming).

    Minimize a function F(X) beginning from a starting point x0, subject to
    optional linear and nonlinear constraints and variable bounds:

    min f(x)
     x

    subject to:

    g(x) = 0            (nonlinear equalities)
    h(x) <= 0           (nonlinear inequalities)
    l <= A*x <= u       (linear constraints)
    xmin <= x <= xmax   (variable bounds)

    Note: The calling syntax is almost identical to that of FMINCON from
    MathWorks' Optimization Toolbox. The main difference is that the linear
    constraints are specified with A, l, u instead of A, B,
    Aeq, Beq. The functions for evaluating the objective function,
    constraints and Hessian are identical.

    Example from http://en.wikipedia.org/wiki/Nonlinear_programming:
    >>> from numpy import array, r_, float64, dot
    >>> from scipy.sparse import csr_matrix
    >>> def f2(x):
    ...     f = -x[0] * x[1] - x[1] * x[2]
    ...     df = -r_[x[1], x[0] + x[2], x[1]]
    ...     # actually not used since 'hess_fcn' is provided
    ...     d2f = -array([[0, 1, 0], [1, 0, 1], [0, 1, 0]], float64)
    ...     return f, df, d2f
    >>> def gh2(x):
    ...     h = dot(array([[1, -1, 1],
    ...                    [1,  1, 1]]), x**2) + array([-2.0, -10.0])
    ...     dh = 2 * csr_matrix(array([[ x[0], x[0]],
    ...                                [-x[1], x[1]],
    ...                                [ x[2], x[2]]]))
    ...     g = array([])
    ...     dg = None
    ...     return h, g, dh, dg
    >>> def hess2(x, lam, cost_mult=1):
    ...     mu = lam["ineqnonlin"]
    ...     a = r_[dot(2 * array([1, 1]), mu), -1, 0]
    ...     b = r_[-1, dot(2 * array([-1, 1]), mu),-1]
    ...     c = r_[0, -1, dot(2 * array([1, 1]), mu)]
    ...     Lxx = csr_matrix(array([a, b, c]))
    ...     return Lxx
    >>> x0 = array([1, 1, 0], float64)
    >>> solution = pips(f2, x0, gh_fcn=gh2, hess_fcn=hess2)
    >>> round(solution["f"], 11) == -7.07106725919
    True
    >>> solution["output"]["iterations"]
    8

    All parameters are optional except f_fcn and x0.

    Parameters:
    ----------
    f_fcn : callable
        Function that evaluates the objective function, its gradients
        and Hessian for a given value of x. If there are
        nonlinear constraints, the Hessian information is provided
        by the 'hess_fcn' argument and is not required here.
    x0 : array, optional
        Starting value of optimization vector x.
    A : csr_matrix, optional
        Optional linear constraints.
    l : array, optional
        Optional linear constraints. Default values are -Inf.
    u : array, optional
        Optional linear constraints. Default values are Inf.
    xmin : array, optional
        Optional lower bounds on the x variables, defaults are
        -Inf.
    xmax : array, optional
        Optional upper bounds on the x variables, defaults are
        Inf.
    gh_fcn : callable, optional
        Function that evaluates the optional nonlinear constraints
        and their gradients for a given value of x.
    hess_fcn : callable, optional
        Handle to function that computes the Hessian of the
        Lagrangian for given values of x, lambda and mu,
        where lambda and mu are the multipliers on the
        equality and inequality constraints, g and h,
        respectively.
    opt : dict, optional
        Optional options dictionary with the following keys, all of
        which are also optional (default values shown in parentheses)
            - 'verbose' (False) - Controls level of progress output
              displayed
            - 'feastol' (1e-6) - termination tolerance for feasibility
              condition
            - 'gradtol' (1e-6) - termination tolerance for gradient
              condition
            - 'comptol' (1e-6) - termination tolerance for
              complementarity condition
            - 'costtol' (1e-6) - termination tolerance for cost
              condition
            - 'max_it' (150) - maximum number of iterations
            - 'step_control' (False) - set to True to enable step-size
              control
            - 'max_red' (20) - maximum number of step-size reductions if
              step-control is on
            - 'cost_mult' (1.0) - cost multiplier used to scale the
              objective function for improved conditioning. Note: This
              value is also passed as the 3rd argument to the Hessian
              evaluation function so that it can appropriately scale the
              objective function term in the Hessian of the Lagrangian.

    Returns:
    --------
    dict
        The solution dictionary has the following keys:
            - 'x' - solution vector
            - 'f' - final objective function value
            - 'converged' - exit status
                - True = first order optimality conditions satisfied
                - False = maximum number of iterations reached
                - None = numerically failed
            - 'output' - output dictionary with keys:
                - 'iterations' - number of iterations performed
                - 'hist' - list of arrays with trajectories of the
                following: feascond, gradcond, compcond, costcond, gamma,
                stepsize, obj, alphap, alphad
                - 'message' - exit message
            - 'lmbda' - dictionary containing the Langrange and Kuhn-Tucker
            multipliers on the constraints, with keys:
                - 'eqnonlin' - nonlinear equality constraints
                - 'ineqnonlin' - nonlinear inequality constraints
                - 'mu_l' - lower (left-hand) limit on linear constraints
                - 'mu_u' - upper (right-hand) limit on linear constraints
                - 'lower' - lower bound on optimization variables
                - 'upper' - upper bound on optimization variables

    Notes
    -----
    1. Ported by Richard Lincoln from the MATLAB Interior Point Solver (MIPS)
        (v1.9) by Ray Zimmerman.  MIPS is distributed as part of the MATPOWER
        project, developed at PSERC, Cornell.

    Reference:

    [1] "On the Computation and Application of Multi-period Security-Constrained
    Optimal Power Flow for Real-time Electricity Market Operations",
    Cornell University, May 2007.

    [2] H. Wang, C. E. Murillo-Sanchez, R. D. Zimmerman, R. J. Thomas, "On Computational
    Issues of Market-Based Optimal Power Flow", IEEE Transactions on Power Systems,
    Vol. 22, No. 3, Aug. 2007, pp. 1185-1193.

    Author
    ------
    Ray Zimmerman (PSERC Cornell)
    """
    if isinstance(f_fcn, dict):  # problem dict
        p = f_fcn
        f_fcn = p['f_fcn']
        x0 = p['x0']
        if 'opt' in p:
            opt = p['opt']
        if 'hess_fcn' in p:
            hess_fcn = p['hess_fcn']
        if 'gh_fcn' in p:
            gh_fcn = p['gh_fcn']
        if 'xmax' in p:
            xmax = p['xmax']
        if 'xmin' in p:
            xmin = p['xmin']
        if 'u' in p:
            u = p['u']
        if 'l' in p:
            l = p['l']
        if 'A' in p:
            A = p['A']

    nx = x0.shape[0]                        # number of variables
    nA = A.shape[0] if A is not None else 0  # number of original linear constr

    # default argument values
    if l is None or len(l) == 0:
        l = -inf * np.ones(nA)
    if u is None or len(u) == 0:
        u = inf * np.ones(nA)
    if xmin is None or len(xmin) == 0:
        xmin = -inf * np.ones(x0.shape[0])
    if xmax is None or len(xmax) == 0:
        xmax = inf * np.ones(x0.shape[0])
    if gh_fcn is None:
        nonlinear = False
        gn = np.array([])
        hn = np.array([])
    else:
        nonlinear = True

    if opt is None:
        opt = {}
    # options
    if "feastol" not in opt:
        opt["feastol"] = 1e-06
    if "gradtol" not in opt:
        opt["gradtol"] = 1e-06
    if "comptol" not in opt:
        opt["comptol"] = 1e-06
    if "costtol" not in opt:
        opt["costtol"] = 1e-06
    if "max_it" not in opt:
        opt["max_it"] = 150
    if "max_red" not in opt:
        opt["max_red"] = 20
    if "step_control" not in opt:
        opt["step_control"] = False
    if "cost_mult" not in opt:
        opt["cost_mult"] = 1
    if "verbose" not in opt:
        opt["verbose"] = 0

    # initialize history
    hist = []

    # constants
    xi = 0.99995
    sigma = 0.1
    z0 = 1
    alpha_min = 1e-8
    rho_min = 0.95
    rho_max = 1.05
    mu_threshold = 1e-5

    # initialize
    i = 0                       # iteration counter
    converged = False           # flag
    eflag = False               # exit flag

    # add var limits to linear constraints
    eyex = sp.eye(nx, nx, format="csr")
    AA = eyex if A is None else sp.vstack([eyex, A], "csr")
    ll = np.r_[xmin, l]
    uu = np.r_[xmax, u]

    # split up linear constraints
    ieq = find(np.abs(uu - ll) <= EPS)
    igt = find((uu >= 1e10) & (ll > -1e10))
    ilt = find((ll <= -1e10) & (uu < 1e10))
    ibx = find((np.abs(uu - ll) > EPS) & (uu < 1e10) & (ll > -1e10))
    # zero-sized sparse matrices unsupported
    Ae = AA[ieq, :] if len(ieq) else None
    if len(ilt) or len(igt) or len(ibx):
        idxs = [(1, ilt), (-1, igt), (1, ibx), (-1, ibx)]
        Ai = sp.vstack([sig * AA[idx, :] for sig, idx in idxs if len(idx)], 'csr')
    else:
        Ai = None
    be = uu[ieq]
    bi = np.r_[uu[ilt], -ll[igt], uu[ibx], -ll[ibx]]

    # evaluate cost f(x0) and constraints g(x0), h(x0)
    x = x0
    f, df = f_fcn(x)                 # cost
    f = f * opt["cost_mult"]
    df = df * opt["cost_mult"]
    if nonlinear:
        hn, gn, dhn, dgn = gh_fcn(x)        # nonlinear constraints
        h = hn if Ai is None else np.r_[hn, Ai * x - bi]  # inequality constraints
        g = gn if Ae is None else np.r_[gn, Ae * x - be]  # equality constraints

        if (dhn is None) and (Ai is None):
            dh = None
        elif dhn is None:
            dh = Ai.T
        elif Ai is None:
            dh = dhn
        else:
            dh = sp.hstack([dhn, Ai.T])

        if (dgn is None) and (Ae is None):
            dg = None
        elif dgn is None:
            dg = Ae.T
        elif Ae is None:
            dg = dgn
        else:
            dg = sp.hstack([dgn, Ae.T])
    else:
        h = -bi if Ai is None else Ai * x - bi        # inequality constraints
        g = -be if Ae is None else Ae * x - be        # equality constraints
        dh = None if Ai is None else Ai.T     # 1st derivative of inequalities
        dg = None if Ae is None else Ae.T     # 1st derivative of equalities

    # some dimensions
    neq = g.shape[0]           # number of equality constraints
    niq = h.shape[0]           # number of inequality constraints
    neqnln = gn.shape[0]       # number of nonlinear equality constraints
    niqnln = hn.shape[0]       # number of nonlinear inequality constraints
    nlt = len(ilt)             # number of upper bounded linear inequalities
    ngt = len(igt)             # number of lower bounded linear inequalities
    nbx = len(ibx)             # number of doubly bounded linear inequalities

    # initialize gamma, lam, mu, z, e
    gamma = 1                  # barrier coefficient
    lam = np.zeros(neq)
    z = z0 * np.ones(niq)
    mu = z0 * np.ones(niq)
    k = find(h < -z0)
    z[k] = -h[k]
    k = find((gamma / z) > z0)
    mu[k] = gamma / z[k]
    e = np.ones(niq)

    # check tolerance
    f0 = f
    if opt["step_control"]:
        L = f + np.dot(lam, g) + np.dot(mu, h + z) - gamma * sum(np.log(z))

    Lx = df.copy()
    Lx = Lx + dg * lam if dg is not None else Lx
    Lx = Lx + dh * mu if dh is not None else Lx

    maxh = np.zeros(1) if len(h) == 0 else max(h)

    gnorm = np.linalg.norm(g, inf) if len(g) else 0.0
    lam_norm = np.linalg.norm(lam, inf) if len(lam) else 0.0
    mu_norm = np.linalg.norm(mu, inf) if len(mu) else 0.0
    znorm = np.linalg.norm(z, inf) if len(z) else 0.0
    feascond = \
        max([gnorm, maxh]) / (1 + max([np.linalg.norm(x, inf), znorm]))
    gradcond = \
        np.linalg.norm(Lx, inf) / (1 + max([lam_norm, mu_norm]))
    compcond = np.dot(z, mu) / (1 + np.linalg.norm(x, inf))
    costcond = np.abs(f - f0) / (1 + np.abs(f0))

    # save history
    hist.append({'feascond': feascond, 'gradcond': gradcond,
                 'compcond': compcond, 'costcond': costcond, 'gamma': gamma,
                 'stepsize': 0, 'obj': f / opt["cost_mult"], 'alphap': 0, 'alphad': 0})

    s = '-sc' if opt["step_control"] else ''
    headers = '  n     objective     step size'
    headers += '    feascond    gradcond'
    headers += '    compcond    costcond'
    head_line = "=== ========  ======  "
    head_line += "   =======  ======="
    head_line += "   =======  ======="
    logger.debug(headers)
    logger.debug(head_line)
    logger.debug("%3d  %12.8g %12g %12g %12g %12g %12g" %
                 (i, (f / opt["cost_mult"]), 0, feascond, gradcond,
                     compcond, costcond))

    if feascond < opt["feastol"] and gradcond < opt["gradtol"] and \
            compcond < opt["comptol"] and costcond < opt["costtol"]:
        converged = True

    # do Newton iterations
    while (not converged) and (i < opt["max_it"]):
        # update iteration counter
        i += 1

        # compute update step
        lmbda = {"eqnonlin": lam[range(neqnln)],
                 "ineqnonlin": mu[range(niqnln)]}
        if nonlinear:
            if hess_fcn is None:
                logger.error("pips: Hessian evaluation via finite differences "
                             "not yet implemented.\nPlease provide "
                             "your own hessian evaluation function.")
            Lxx = hess_fcn(x, lmbda, opt["cost_mult"])
        else:
            _, _, d2f = f_fcn(x, True)      # cost
            Lxx = d2f * opt["cost_mult"]
        rz = range(len(z))
        zinvdiag = c_sparse((1.0 / z, (rz, rz))) if len(z) else None
        rmu = range(len(mu))
        mudiag = c_sparse((mu, (rmu, rmu))) if len(mu) else None
        dh_zinv = None if dh is None else dh * zinvdiag
        M = Lxx if dh is None else Lxx + dh_zinv * mudiag * dh.T
        N = Lx if dh is None else Lx + dh_zinv * (mudiag * h + gamma * e)

        Ab = c_sparse(M) if dg is None else sp.vstack([
            sp.hstack([M, dg]),
            sp.hstack([dg.T, c_sparse((neq, neq))])
        ])
        bb = np.r_[-N, -g]

        dxdlam = sp.linalg.spsolve(Ab.tocsr(), bb)

        if np.any(np.isnan(dxdlam)):
            print('\nNumerically Failed\n')
            eflag = -1
            break

        dx = dxdlam[:nx]
        dlam = dxdlam[nx:nx + neq]
        dz = -h - z if dh is None else -h - z - dh.T * dx
        dmu = -mu if dh is None else -mu + zinvdiag * (gamma * e - mudiag * dz)

        # optional step-size control
        sc = False
        if opt["step_control"]:
            x1 = x + dx

            # evaluate cost, constraints, derivatives at x1
            f1, df1 = f_fcn(x1)          # cost
            f1 = f1 * opt["cost_mult"]
            df1 = df1 * opt["cost_mult"]
            if nonlinear:
                hn1, gn1, dhn1, dgn1 = gh_fcn(x1)  # nonlinear constraints

                h1 = hn1 if Ai is None else np.r_[hn1, Ai * x1 - bi]  # ieq constraints
                g1 = gn1 if Ae is None else np.r_[gn1, Ae * x1 - be]  # eq constraints

                # 1st der of ieq
                if (dhn1 is None) and (Ai is None):
                    dh1 = None
                elif dhn1 is None:
                    dh1 = Ai.T
                elif Ai is None:
                    dh1 = dhn1
                else:
                    dh1 = sp.hstack([dhn1, Ai.T])

                # 1st der of eqs
                if (dgn1 is None) and (Ae is None):
                    dg1 = None
                elif dgn is None:
                    dg1 = Ae.T
                elif Ae is None:
                    dg1 = dgn1
                else:
                    dg1 = sp.hstack([dgn1, Ae.T])
            else:
                h1 = -bi if Ai is None else Ai * x1 - bi    # inequality constraints
                g1 = -be if Ae is None else Ae * x1 - be    # equality constraints

                dh1 = dh  # 1st derivative of inequalities
                dg1 = dg  # 1st derivative of equalities

            # check tolerance
            Lx1 = df1
            Lx1 = Lx1 + dg1 * lam if dg1 is not None else Lx1
            Lx1 = Lx1 + dh1 * mu if dh1 is not None else Lx1

            maxh1 = np.zeros(1) if len(h1) == 0 else max(h1)

            g1norm = np.linalg.norm(g1, inf) if len(g1) else 0.0
            lam1_norm = np.linalg.norm(lam, inf) if len(lam) else 0.0
            mu1_norm = np.linalg.norm(mu, inf) if len(mu) else 0.0
            z1norm = np.linalg.norm(z, inf) if len(z) else 0.0

            feascond1 = max([g1norm, maxh1]) / \
                (1 + max([np.linalg.norm(x1, inf), z1norm]))
            gradcond1 = np.linalg.norm(Lx1, inf) / (1 + max([lam1_norm, mu1_norm]))

            if (feascond1 > feascond) and (gradcond1 > gradcond):
                sc = True
        if sc:
            alpha = 1.0
            for j in range(opt["max_red"]):
                dx1 = alpha * dx
                x1 = x + dx1
                f1, _ = f_fcn(x1)             # cost
                f1 = f1 * opt["cost_mult"]
                if nonlinear:
                    hn1, gn1, _, _ = gh_fcn(x1)              # nonlinear constraints
                    h1 = hn1 if Ai is None else np.r_[hn1, Ai * x1 - bi]         # inequality constraints
                    g1 = gn1 if Ae is None else np.r_[gn1, Ae * x1 - be]         # equality constraints
                else:
                    h1 = -bi if Ai is None else Ai * x1 - bi    # inequality constraints
                    g1 = -be if Ae is None else Ae * x1 - be    # equality constraints

                L1 = f1 + np.dot(lam, g1) + np.dot(mu, h1 + z) - gamma * sum(np.log(z))

                logger.info("   %3d            %10.5f" % (-j, np.linalg.norm(dx1)))

                rho = (L1 - L) / (np.dot(Lx, dx1) + 0.5 * np.dot(dx1, Lxx * dx1))

                if (rho > rho_min) and (rho < rho_max):
                    break
                else:
                    alpha = alpha / 2.0
            dx = alpha * dx
            dz = alpha * dz
            dlam = alpha * dlam
            dmu = alpha * dmu

        # do the update
        k = find(dz < 0.0)
        alphap = min([xi * min(z[k] / -dz[k]), 1]) if len(k) else 1.0
        k = find(dmu < 0.0)
        alphad = min([xi * min(mu[k] / -dmu[k]), 1]) if len(k) else 1.0
        x = x + alphap * dx
        z = z + alphap * dz
        lam = lam + alphad * dlam
        mu = mu + alphad * dmu
        if niq > 0:
            gamma = sigma * np.dot(z, mu) / niq

        # evaluate cost, constraints, derivatives
        f, df = f_fcn(x)             # cost
        f = f * opt["cost_mult"]
        df = df * opt["cost_mult"]
        if nonlinear:
            hn, gn, dhn, dgn = gh_fcn(x)                   # nln constraints
            h = hn if Ai is None else np.r_[hn, Ai * x - bi]  # ieq constr
            g = gn if Ae is None else np.r_[gn, Ae * x - be]  # eq constr

            if (dhn is None) and (Ai is None):
                dh = None
            elif dhn is None:
                dh = Ai.T
            elif Ai is None:
                dh = dhn
            else:
                dh = sp.hstack([dhn, Ai.T])

            if (dgn is None) and (Ae is None):
                dg = None
            elif dgn is None:
                dg = Ae.T
            elif Ae is None:
                dg = dgn
            else:
                dg = sp.hstack([dgn, Ae.T])
        else:
            h = -bi if Ai is None else Ai * x - bi    # inequality constraints
            g = -be if Ae is None else Ae * x - be    # equality constraints
            # 1st derivatives are constant, still dh = Ai.T, dg = Ae.T

        Lx = df
        Lx = Lx + dg * lam if dg is not None else Lx
        Lx = Lx + dh * mu if dh is not None else Lx

        if len(h) == 0:
            maxh = np.zeros(1)
        else:
            maxh = max(h)

        gnorm = np.linalg.norm(g, inf) if len(g) else 0.0
        lam_norm = np.linalg.norm(lam, inf) if len(lam) else 0.0
        mu_norm = np.linalg.norm(mu, inf) if len(mu) else 0.0
        znorm = np.linalg.norm(z, inf) if len(z) else 0.0
        feascond = \
            max([gnorm, maxh]) / (1 + max([np.linalg.norm(x, inf), znorm]))
        gradcond = \
            np.linalg.norm(Lx, inf) / (1 + max([lam_norm, mu_norm]))
        compcond = np.dot(z, mu) / (1 + np.linalg.norm(x, inf))
        costcond = float(np.abs(f - f0) / (1 + np.abs(f0)))

        hist.append({'feascond': feascond, 'gradcond': gradcond,
                     'compcond': compcond, 'costcond': costcond, 'gamma': gamma,
                     'stepsize': np.linalg.norm(dx), 'obj': f / opt["cost_mult"],
                     'alphap': alphap, 'alphad': alphad})

        logger.debug("%3d  %12.8g %10.5g %12g %12g %12g %12g" %
                     (i, (f / opt["cost_mult"]), np.linalg.norm(dx), feascond, gradcond,
                      compcond, costcond))

        if feascond < opt["feastol"] and gradcond < opt["gradtol"] and \
                compcond < opt["comptol"] and costcond < opt["costtol"]:
            converged = True
        else:
            if np.any(np.isnan(x)) or (alphap < alpha_min) or \
                    (alphad < alpha_min) or (gamma < EPS) or (gamma > 1.0 / EPS):
                eflag = -1
                break
            f0 = f

            if opt["step_control"]:
                L = f + np.dot(lam, g) + np.dot(mu, (h + z)) - gamma * sum(np.log(z))

    if not converged:
        logger.debug("Did not converge in %d iterations." % i)

    # package results
    if eflag != -1:
        eflag = converged

    if eflag == 0:
        message = 'Did not converge'
    elif eflag == 1:
        message = 'Converged'
    elif eflag == -1:
        message = 'Numerically failed'
    else:
        raise

    output = {"iterations": i, "hist": hist, "message": message}

    # zero out multipliers on non-binding constraints
    mu[find((h < -opt["feastol"]) & (mu < mu_threshold))] = 0.0

    # un-scale cost and prices
    f = f / opt["cost_mult"]
    lam = lam / opt["cost_mult"]
    mu = mu / opt["cost_mult"]

    # re-package multipliers into struct
    lam_lin = lam[neqnln:neq]           # lambda for linear constraints
    mu_lin = mu[niqnln:niq]             # mu for linear constraints
    kl = find(lam_lin < 0.0)     # lower bound binding
    ku = find(lam_lin > 0.0)     # upper bound binding

    mu_l = np.zeros(nx + nA)
    mu_l[ieq[kl]] = -lam_lin[kl]
    mu_l[igt] = mu_lin[nlt:nlt + ngt]
    mu_l[ibx] = mu_lin[nlt + ngt + nbx:nlt + ngt + nbx + nbx]

    mu_u = np.zeros(nx + nA)
    mu_u[ieq[ku]] = lam_lin[ku]
    mu_u[ilt] = mu_lin[:nlt]
    mu_u[ibx] = mu_lin[nlt + ngt:nlt + ngt + nbx]

    lmbda = {'mu_l': mu_l[nx:], 'mu_u': mu_u[nx:],
             'lower': mu_l[:nx], 'upper': mu_u[:nx]}

    if niqnln > 0:
        lmbda['ineqnonlin'] = mu[:niqnln]
    if neqnln > 0:
        lmbda['eqnonlin'] = lam[:neqnln]

#    lmbda = {"eqnonlin": lam[:neqnln], 'ineqnonlin': mu[:niqnln],
#             "mu_l": mu_l[nx:], "mu_u": mu_u[nx:],
#             "lower": mu_l[:nx], "upper": mu_u[:nx]}

    solution = {"x": x, "f": f, "eflag": converged,
                "output": output, "lmbda": lmbda}

    return solution


def pipsver(*args):
    """
    Return PIPS version information for the current installation.

    Author: Ray Zimmerman (PSERC Cornell)

    Returns
    -------
    dict
        A dictionary containing PIPS version information with the following keys:
        - 'Name': Name of the software (PIPS)
        - 'Version': Version number (1.0)
        - 'Release': Release information (empty string)
        - 'Date': Date of release (07-Feb-2011)

    Author
    ------
    Ray Zimmerman (PSERC Cornell)
    """
    ver = {'Name': 'PIPS',
           'Version': '1.0',
           'Release':  '',
           'Date': '07-Feb-2011'}

    return ver


def pipsopf_solver(om, ppopt, out_opt=None):
    """
    Solves AC optimal power flow using PIPS.

    Inputs are an OPF model object, a PYPOWER options vector, and
    a dict containing keys (can be empty) for each of the desired
    optional output fields.

    Outputs are a `results` dict, a `success` flag, and a `raw` output dict.

    Parameters
    ----------
    om : object
        OPF model object.
    ppopt : dict
        PYPOWER options vector.
    out_opt : dict, optional
        Dictionary containing keys for optional output fields (default is None).

    Returns
    -------
    dict
        A `results` dictionary containing various fields, including optimization
        variables, objective function value, and shadow prices on constraints.
    bool
        A `success` flag indicating whether the solver converged successfully.
    dict
        A `raw` output dictionary in the form returned by MINOS, containing
        information about optimization variables, constraint multipliers, solver
        termination code, and solver-specific output information.

    Notes
    -----
    The `results` dictionary contains fields such as 'order', 'x', 'f', 'mu',
    where 'mu' includes shadow prices on variables and constraints.

    The `success` flag is `True` if the solver converged successfully, and `False`
    otherwise.

    The `raw` output dictionary includes information specific to the solver.

    Author
    ------
    Ray Zimmerman (PSERC Cornell)

    Carlos E. Murillo-Sanchez (PSERC Cornell & Universidad Autonoma de Manizales)
    """
    # ----- initialization -----
    # optional output
    if out_opt is None:
        out_opt = {}

    # options
    verbose = ppopt['VERBOSE']
    feastol = ppopt['PDIPM_FEASTOL']
    gradtol = ppopt['PDIPM_GRADTOL']
    comptol = ppopt['PDIPM_COMPTOL']
    costtol = ppopt['PDIPM_COSTTOL']
    max_it = ppopt['PDIPM_MAX_IT']
    max_red = ppopt['SCPDIPM_RED_IT']
    step_control = (ppopt['OPF_ALG'] == 565)  # OPF_ALG == 565, PIPS-sc
    if feastol == 0:
        feastol = ppopt['OPF_VIOLATION']
    opt = {'feastol': feastol,
           'gradtol': gradtol,
           'comptol': comptol,
           'costtol': costtol,
           'max_it': max_it,
           'max_red': max_red,
           'step_control': step_control,
           'cost_mult': 1e-4,
           'verbose': verbose}

    # unpack data
    ppc = om.get_ppc()
    baseMVA, bus, gen, branch, gencost = \
        ppc["baseMVA"], ppc["bus"], ppc["gen"], ppc["branch"], ppc["gencost"]
    vv, _, nn, _ = om.get_idx()

    # problem dimensions
    nb = bus.shape[0]  # number of buses
    nl = branch.shape[0]  # number of branches
    ny = om.getN('var', 'y')  # number of piece-wise linear costs

    # linear constraints
    A, l, u = om.linear_constraints()

    # bounds on optimization vars
    _, xmin, xmax = om.getv()

    # build admittance matrices
    Ybus, Yf, Yt = makeYbus(baseMVA, bus, branch)

    # try to select an interior initial point
    ll, uu = xmin.copy(), xmax.copy()
    ll[xmin == -inf] = -1e10  # replace Inf with numerical proxies
    uu[xmax == inf] = 1e10
    x0 = (ll + uu) / 2
    Varefs = bus[bus[:, IDX.bus.BUS_TYPE] == IDX.bus.REF, IDX.bus.VA] * deg2rad
    # angles set to first reference angle
    x0[vv["i1"]["Va"]:vv["iN"]["Va"]] = Varefs[0]
    if ny > 0:
        ipwl = find(gencost[:, IDX.cost.MODEL] == IDX.cost.PW_LINEAR)
#         PQ = np.r_[gen[:, PMAX], gen[:, QMAX]]
#         c = totcost(gencost[ipwl, :], PQ[ipwl])
        # largest y-value in CCV data
        c = gencost.flatten('F')[sub2ind(gencost.shape, ipwl, IDX.cost.NCOST+2*gencost[ipwl, IDX.cost.NCOST])]
        x0[vv["i1"]["y"]:vv["iN"]["y"]] = max(c) + 0.1 * abs(max(c))
#        x0[vv["i1"]["y"]:vv["iN"]["y"]] = c + 0.1 * abs(c)

    # find branches with flow limits
    il = find((branch[:, IDX.branch.RATE_A] != 0) & (branch[:, IDX.branch.RATE_A] < 1e10))
    nl2 = len(il)  # number of constrained lines

    # -----  run opf  -----
    def f_fcn(x, return_hessian=False): return opf_costfcn(x, om, return_hessian)
    def gh_fcn(x): return opf_consfcn(x, om, Ybus, Yf[il, :], Yt[il, :], ppopt, il)

    def hess_fcn(
        x, lmbda, cost_mult): return opf_hessfcn(
            x, lmbda, om, Ybus, Yf[il, :],
            Yt[il, :],
        ppopt, il, cost_mult)

    solution = pips(f_fcn, x0, A, l, u, xmin, xmax, gh_fcn, hess_fcn, opt)
    x, f, info, lmbda, output = solution["x"], solution["f"], \
        solution["eflag"], solution["lmbda"], solution["output"]

    success = (info > 0)

    # update solution data
    Va = x[vv["i1"]["Va"]:vv["iN"]["Va"]]
    Vm = x[vv["i1"]["Vm"]:vv["iN"]["Vm"]]
    Pg = x[vv["i1"]["Pg"]:vv["iN"]["Pg"]]
    Qg = x[vv["i1"]["Qg"]:vv["iN"]["Qg"]]

    V = Vm * np.exp(1j * Va)

    # -----  calculate return values  -----
    # update voltages & generator outputs
    bus[:, IDX.bus.VA] = Va * rad2deg

    bus[:, IDX.bus.VM] = Vm
    gen[:, IDX.gen.PG] = Pg * baseMVA
    gen[:, IDX.gen.QG] = Qg * baseMVA
    gen[:, IDX.gen.VG] = Vm[gen[:, IDX.gen.GEN_BUS].astype(int)]

    # compute branch flows
    Sf = V[branch[:, IDX.branch.F_BUS].astype(int)] * np.conj(Yf * V)  # cplx pwr at "from" bus, p["u"].
    St = V[branch[:, IDX.branch.T_BUS].astype(int)] * np.conj(Yt * V)  # cplx pwr at "to" bus, p["u"].
    branch[:, IDX.branch.PF] = Sf.real * baseMVA
    branch[:, IDX.branch.QF] = Sf.imag * baseMVA
    branch[:, IDX.branch.PT] = St.real * baseMVA
    branch[:, IDX.branch.QT] = St.imag * baseMVA

    # line constraint is actually on square of limit
    # so we must fix multipliers
    muSf = np.zeros(nl)
    muSt = np.zeros(nl)
    if len(il) > 0:
        muSf[il] = \
            2 * lmbda["ineqnonlin"][:nl2] * branch[il, IDX.branch.RATE_A] / baseMVA
        muSt[il] = \
            2 * lmbda["ineqnonlin"][nl2:nl2+nl2] * branch[il, IDX.branch.RATE_A] / baseMVA

    # update Lagrange multipliers
    bus[:, IDX.bus.MU_VMAX] = lmbda["upper"][vv["i1"]["Vm"]:vv["iN"]["Vm"]]
    bus[:, IDX.bus.MU_VMIN] = lmbda["lower"][vv["i1"]["Vm"]:vv["iN"]["Vm"]]
    gen[:, IDX.gen.MU_PMAX] = lmbda["upper"][vv["i1"]["Pg"]:vv["iN"]["Pg"]] / baseMVA
    gen[:, IDX.gen.MU_PMIN] = lmbda["lower"][vv["i1"]["Pg"]:vv["iN"]["Pg"]] / baseMVA
    gen[:, IDX.gen.MU_QMAX] = lmbda["upper"][vv["i1"]["Qg"]:vv["iN"]["Qg"]] / baseMVA
    gen[:, IDX.gen.MU_QMIN] = lmbda["lower"][vv["i1"]["Qg"]:vv["iN"]["Qg"]] / baseMVA

    bus[:, IDX.bus.LAM_P] = \
        lmbda["eqnonlin"][nn["i1"]["Pmis"]:nn["iN"]["Pmis"]] / baseMVA
    bus[:, IDX.bus.LAM_Q] = \
        lmbda["eqnonlin"][nn["i1"]["Qmis"]:nn["iN"]["Qmis"]] / baseMVA
    branch[:, IDX.branch.MU_SF] = muSf / baseMVA
    branch[:, IDX.branch.MU_ST] = muSt / baseMVA

    # package up results
    nlnN = om.getN('nln')

    # extract multipliers for nonlinear constraints
    kl = find(lmbda["eqnonlin"] < 0)
    ku = find(lmbda["eqnonlin"] > 0)
    nl_mu_l = np.zeros(nlnN)
    nl_mu_u = np.r_[np.zeros(2*nb), muSf, muSt]
    nl_mu_l[kl] = -lmbda["eqnonlin"][kl]
    nl_mu_u[ku] = lmbda["eqnonlin"][ku]

    mu = {
        'var': {'l': lmbda["lower"], 'u': lmbda["upper"]},
        'nln': {'l': nl_mu_l, 'u': nl_mu_u},
        'lin': {'l': lmbda["mu_l"], 'u': lmbda["mu_u"]}}

    results = ppc
    results["bus"], results["branch"], results["gen"], \
        results["om"], results["x"], results["mu"], results["f"] = \
        bus, branch, gen, om, x, mu, f

    pimul = np.r_[
        results["mu"]["nln"]["l"] - results["mu"]["nln"]["u"],
        results["mu"]["lin"]["l"] - results["mu"]["lin"]["u"],
        -np.ones(int(ny > 0)),
        results["mu"]["var"]["l"] - results["mu"]["var"]["u"],
    ]
    raw = {'xr': x, 'pimul': pimul, 'info': info, 'output': output}

    return results, success, raw



"""
DCOPF routines.
"""
import logging

import numpy as np
from ams.core.param import RParam
from ams.core.service import NumOp, NumOpDual

from ams.routines.dcpf import DCPFBase

from ams.opt import Constraint, Objective, ExpressionCalc, Expression


logger = logging.getLogger(__name__)


class DCOPF(DCPFBase):
    """
    DC optimal power flow (DCOPF).

    Notes
    -----
    1. The nodal price is calculated as ``pi`` in ``pic``.
    1. Devices online status of `StaticGen`, `StaticLoad`, and `Shunt` are considered in the connectivity
    matrices `Cft`, `Cg`, `Cl`, and `Csh`.

    References
    ----------
    1. R. D. Zimmerman, C. E. Murillo-Sanchez, and R. J. Thomas, “MATPOWER: Steady-State Operations, Planning, and
    Analysis Tools for Power Systems Research and Education,” IEEE Trans. Power Syst., vol. 26, no. 1, pp. 12–19,
    Feb. 2011
    """

    def __init__(self, system, config):
        DCPFBase.__init__(self, system, config)
        self.info = 'DC Optimal Power Flow'
        self.type = 'DCED'

        # --- Mapping Section ---
        # --- from map ---
        self.map1.update({
            'ug': ('StaticGen', 'u'),
        })
        # --- to map ---
        self.map2.update({
            'vBus': ('Bus', 'v0'),
            'ug': ('StaticGen', 'u'),
            'pg': ('StaticGen', 'p0'),
        })

        # --- Data Section ---
        # --- generator cost ---
        self.c2 = RParam(info='Gen cost coefficient 2',
                         name='c2', tex_name=r'c_{2}',
                         unit=r'$/(p.u.^2)', model='GCost',
                         indexer='gen', imodel='StaticGen',
                         nonneg=True, no_parse=True)
        self.c1 = RParam(info='Gen cost coefficient 1',
                         name='c1', tex_name=r'c_{1}',
                         unit=r'$/(p.u.)', model='GCost',
                         indexer='gen', imodel='StaticGen',)
        self.c0 = RParam(info='Gen cost coefficient 0',
                         name='c0', tex_name=r'c_{0}',
                         unit=r'$', model='GCost',
                         indexer='gen', imodel='StaticGen',
                         no_parse=True)
        # --- generator ---
        self.ctrl = RParam(info='Gen controllability',
                           name='ctrl', tex_name=r'c_{trl}',
                           model='StaticGen', src='ctrl',
                           no_parse=True)
        self.ctrle = NumOpDual(info='Effective Gen controllability',
                               name='ctrle', tex_name=r'c_{trl, e}',
                               u=self.ctrl, u2=self.ug,
                               fun=np.multiply, no_parse=True)
        self.nctrl = NumOp(u=self.ctrl, fun=np.logical_not,
                           name='nctrl', tex_name=r'c_{trl,n}',
                           info='Effective Gen uncontrollability',
                           no_parse=True,)
        self.nctrle = NumOpDual(info='Effective Gen uncontrollability',
                                name='nctrle', tex_name=r'c_{trl,n,e}',
                                u=self.nctrl, u2=self.ug,
                                fun=np.multiply, no_parse=True)
        self.pmax = RParam(info='Gen maximum active power',
                           name='pmax', tex_name=r'p_{g, max}',
                           unit='p.u.', model='StaticGen',
                           no_parse=False,)
        self.pmin = RParam(info='Gen minimum active power',
                           name='pmin', tex_name=r'p_{g, min}',
                           unit='p.u.', model='StaticGen',
                           no_parse=False,)

        # --- line ---
        self.ul = RParam(info='Line connection status',
                         name='ul', tex_name=r'u_{l}',
                         model='Line', src='u',
                         no_parse=True,)
        self.rate_a = RParam(info='long-term flow limit',
                             name='rate_a', tex_name=r'R_{ATEA}',
                             unit='p.u.', model='Line',)
        # --- line angle difference ---
        self.amax = RParam(model='Line', src='amax',
                           name='amax', tex_name=r'\theta_{bus, max}',
                           info='max line angle difference',
                           no_parse=True,)
        self.amin = RParam(model='Line', src='amin',
                           name='amin', tex_name=r'\theta_{bus, min}',
                           info='min line angle difference',
                           no_parse=True,)

        # --- Model Section ---
        self.pmaxe = Expression(info='Effective pmax',
                                name='pmaxe', tex_name=r'p_{g, max, e}',
                                e_str='mul(nctrle, pg0) + mul(ctrle, pmax)',
                                model='StaticGen', src=None, unit='p.u.',)
        self.pmine = Expression(info='Effective pmin',
                                name='pmine', tex_name=r'p_{g, min, e}',
                                e_str='mul(nctrle, pg0) + mul(ctrle, pmin)',
                                model='StaticGen', src=None, unit='p.u.',)
        self.pglb = Constraint(name='pglb', info='pg min',
                               e_str='-pg + pmine', is_eq=False,)
        self.pgub = Constraint(name='pgub', info='pg max',
                               e_str='pg - pmaxe', is_eq=False,)

        # --- line flow ---
        self.plflb = Constraint(info='line flow lower bound',
                                name='plflb', is_eq=False,
                                e_str='-plf - mul(ul, rate_a)',)
        self.plfub = Constraint(info='line flow upper bound',
                                name='plfub', is_eq=False,
                                e_str='plf - mul(ul, rate_a)',)
        self.alflb = Constraint(info='line angle difference lower bound',
                                name='alflb', is_eq=False,
                                e_str='-CftT@aBus + amin',)
        self.alfub = Constraint(info='line angle difference upper bound',
                                name='alfub', is_eq=False,
                                e_str='CftT@aBus - amax',)
        # NOTE: in CVXPY, dual_variables returns a list
        self.pi = ExpressionCalc(info='LMP, dual of <pb>',
                                 name='pi', unit='$/p.u.',
                                 model='Bus', src=None,
                                 e_str='pb.dual_variables[0]')

        # --- objective ---
        obj = 'sum(mul(c2, pg**2))'
        obj += '+ sum(mul(c1, pg))'
        obj += '+ sum(mul(ug, c0))'
        self.obj = Objective(name='obj',
                             info='total cost', unit='$',
                             sense='min', e_str=obj,)

    def dc2ac(self, kloss=1.0, **kwargs):
        """
        Convert the RTED results with ACOPF.

        Parameters
        ----------
        kloss : float, optional
            The loss factor for the conversion. Defaults to 1.0.
        """
        exec_time = self.exec_time
        if self.exec_time == 0 or self.exit_code != 0:
            logger.warning(f'{self.class_name} is not executed successfully, quit conversion.')
            return False

        # --- ACOPF ---
        # scale up load
        pq_idx = self.system.StaticLoad.get_all_idxes()
        pd0 = self.system.StaticLoad.get(src='p0', attr='v', idx=pq_idx).copy()
        qd0 = self.system.StaticLoad.get(src='q0', attr='v', idx=pq_idx).copy()
        self.system.StaticLoad.set(src='p0', idx=pq_idx, attr='v', value=pd0 * kloss)
        self.system.StaticLoad.set(src='q0', idx=pq_idx, attr='v', value=qd0 * kloss)
        # run ACOPF
        ACOPF = self.system.ACOPF
        ACOPF.run()
        # scale load back
        self.system.StaticLoad.set(src='p0', idx=pq_idx, attr='v', value=pd0)
        self.system.StaticLoad.set(src='q0', idx=pq_idx, attr='v', value=qd0)
        if not ACOPF.exit_code == 0:
            logger.warning('<ACOPF> did not converge, conversion failed.')
            # NOTE: mock results to fit interface with ANDES
            self.vBus = ACOPF.vBus
            self.vBus.optz.value = np.ones(self.system.Bus.n)
            self.aBus.optz.value = np.zeros(self.system.Bus.n)
            return False
        self.pg.optz.value = ACOPF.pg.v

        # NOTE: mock results to fit interface with ANDES
        self.vBus.optz.value = ACOPF.vBus.v
        self.aBus.optz.value = ACOPF.aBus.v
        self.exec_time = exec_time

        # --- set status ---
        self.system.recent = self
        self.converted = True
        logger.warning(f'<{self.class_name}> converted to AC.')
        return True

    def run(self, **kwargs):
        """
        Run the routine.

        Following kwargs go to `self.init()`: `force`, `force_mats`, `force_constr`, `force_om`.

        Following kwargs go to `self.solve()`: `solver`, `verbose`, `gp`, `qcp`, `requires_grad`,
        `enforce_dpp`, `ignore_dpp`, `method`, and all rest.

        Parameters
        ----------
        force : bool, optional
            If True, force re-initialization. Defaults to False.
        force_mats : bool, optional
            If True, force re-generating matrices. Defaults to False.
        force_constr : bool, optional
            Whether to turn on all constraints.
        force_om : bool, optional
            If True, force re-generating optimization model. Defaults to False.
        solver: str, optional
            The solver to use. For example, 'GUROBI', 'ECOS', 'SCS', or 'OSQP'.
        verbose : bool, optional
            Overrides the default of hiding solver output and prints logging
            information describing CVXPY's compilation process.
        gp : bool, optional
            If True, parses the problem as a disciplined geometric program
            instead of a disciplined convex program.
        qcp : bool, optional
            If True, parses the problem as a disciplined quasiconvex program
            instead of a disciplined convex program.
        requires_grad : bool, optional
            Makes it possible to compute gradients of a solution with respect to Parameters
            by calling problem.backward() after solving, or to compute perturbations to the variables
            given perturbations to Parameters by calling problem.derivative().
            Gradients are only supported for DCP and DGP problems, not quasiconvex problems.
            When computing gradients (i.e., when this argument is True), the problem must satisfy the DPP rules.
        enforce_dpp : bool, optional
            When True, a DPPError will be thrown when trying to solve a
            non-DPP problem (instead of just a warning).
            Only relevant for problems involving Parameters. Defaults to False.
        ignore_dpp : bool, optional
            When True, DPP problems will be treated as non-DPP, which may speed up compilation. Defaults to False.
        method : function, optional
            A custom solve method to use.
        """
        return super().run(**kwargs)



import unittest

import ams


class TestOModelWrappers(unittest.TestCase):
    """
    Test wrappers in module `omodel`.
    """

    def setUp(self) -> None:
        self.ss = ams.load(ams.get_case("matpower/case5.m"),
                           setup=True,
                           default_config=True,
                           no_output=True,
                           )

    def test_ensure_symbols(self):
        """
        Test `ensure_symbols` wrapper.
        """
        self.assertFalse(self.ss.DCOPF._syms, "Symbols should not be ready before generation!")
        self.ss.DCOPF.pd.parse()
        self.assertTrue(self.ss.DCOPF._syms, "Symbols should be ready after generation!")

    def test_ensure_mats_and_parsed(self):
        """
        Test `ensure_mats_and_parsed` wrapper.
        """
        self.assertFalse(self.ss.mats.initialized, "MatProcessor should not be initialized before initialization!")
        self.assertFalse(self.ss.DCOPF.om.parsed, "OModel should be parsed after parsing!")
        self.ss.DCOPF.pd.evaluate()
        self.assertTrue(self.ss.mats.initialized, "MatProcessor should be initialized after parsing!")
        self.assertTrue(self.ss.DCOPF.om.parsed, "OModle should be parsed after parsing!")


class TestOModel(unittest.TestCase):
    """
    Test class `OModel`.
    """

    def setUp(self) -> None:
        self.ss = ams.load(ams.get_case("matpower/case5.m"),
                           setup=True,
                           default_config=True,
                           no_output=True,
                           )

    def test_initialized(self):
        """
        Test `OModel` initialization.
        """
        self.assertFalse(self.ss.DCOPF.om.initialized, "OModel shoud not be initialized before initialziation!")
        self.ss.DCOPF.om.init()
        self.assertTrue(self.ss.DCOPF.om.initialized, "OModel should be initialized!")

    def test_uninitialized_after_nonparametric_update(self):
        """
        Test `OModel` initialization after nonparametric update.
        """
        self.ss.DCOPF.om.init()
        self.assertTrue(self.ss.DCOPF.om.initialized, "OModel should be initialized after initialization!")
        self.ss.DCOPF.update('ug')
        self.assertTrue(self.ss.DCOPF.om.initialized, "OModel should be initialized after nonparametric update!")


class TestOModelrepr(unittest.TestCase):
    """
    Test __repr__ in module `omodel`.
    """

    def setUp(self) -> None:
        self.ss = ams.load(ams.get_case("5bus/pjm5bus_demo.xlsx"),
                           setup=True,
                           default_config=True,
                           no_output=True,
                           )

    def test_dcopf(self):
        """
        Test `DCOPF.om` __repr__().
        """
        self.ss.DCOPF.om.parse()
        self.assertIsInstance(self.ss.DCOPF.pg.__repr__(), str)
        self.assertIsInstance(self.ss.DCOPF.pmax.__repr__(), str)
        self.assertIsInstance(self.ss.DCOPF.pb.__repr__(), str)
        self.assertIsInstance(self.ss.DCOPF.plflb.__repr__(), str)
        self.assertIsInstance(self.ss.DCOPF.obj.__repr__(), str)

    def test_rted(self):
        """
        Test `RTED.om` __repr__().
        """
        self.ss.RTED.om.parse()
        self.assertIsInstance(self.ss.RTED.pru.__repr__(), str)
        self.assertIsInstance(self.ss.RTED.R10.__repr__(), str)
        self.assertIsInstance(self.ss.RTED.pru.__repr__(), str)
        self.assertIsInstance(self.ss.RTED.rru.__repr__(), str)
        self.assertIsInstance(self.ss.RTED.obj.__repr__(), str)

    def test_ed(self):
        """
        Test `ED.om` __repr__().
        """
        self.ss.ED.om.parse()
        self.assertIsInstance(self.ss.ED.prs.__repr__(), str)
        self.assertIsInstance(self.ss.ED.R30.__repr__(), str)
        self.assertIsInstance(self.ss.ED.prsb.__repr__(), str)
        self.assertIsInstance(self.ss.ED.obj.__repr__(), str)

    def test_uc(self):
        """
        Test `UC.om` __repr__().
        """
        self.ss.UC.om.parse()
        self.assertIsInstance(self.ss.UC.prns.__repr__(), str)
        self.assertIsInstance(self.ss.UC.cdp.__repr__(), str)
        self.assertIsInstance(self.ss.UC.prnsb.__repr__(), str)
        self.assertIsInstance(self.ss.UC.obj.__repr__(), str)



"""
DC power flow routines using PYPOWER.
"""
import logging

from andes.shared import deg2rad
from andes.utils.misc import elapsed

from ams.routines.routine import RoutineBase
from ams.opt import Var
from ams.pypower import runpf
from ams.pypower.core import ppoption

from ams.io.pypower import system2ppc
from ams.core.param import RParam

logger = logging.getLogger(__name__)


class DCPF0(RoutineBase):
    """
    DC power flow using PYPOWER.

    This class is deprecated as of version 0.9.12 and will be removed in 1.1.0.

    Notes
    -----
    1. DCPF is solved with PYPOWER ``runpf`` function.
    2. DCPF formulation is not complete yet, but this does not affect the
       results because the data are passed to PYPOWER for solving.
    """

    def __init__(self, system, config):
        RoutineBase.__init__(self, system, config)
        self.info = 'DC Power Flow'
        self.type = 'PF'

        self.ug = RParam(info='Gen connection status',
                         name='ug', tex_name=r'u_{g}',
                         model='StaticGen', src='u',
                         no_parse=True)

        # --- routine data ---
        self.x = RParam(info="line reactance",
                        name='x', tex_name='x',
                        unit='p.u.',
                        model='Line', src='x',)
        self.tap = RParam(info="transformer branch tap ratio",
                          name='tap', tex_name=r't_{ap}',
                          model='Line', src='tap',
                          unit='float',)
        self.phi = RParam(info="transformer branch phase shift in rad",
                          name='phi', tex_name=r'\phi',
                          model='Line', src='phi',
                          unit='radian',)

        # --- load ---
        self.pd = RParam(info='active deman',
                         name='pd', tex_name=r'p_{d}',
                         unit='p.u.',
                         model='StaticLoad', src='p0')
        # --- gen ---
        self.pg = Var(info='Gen active power',
                      unit='p.u.',
                      name='pg', tex_name=r'p_{g}',
                      model='StaticGen', src='p',)

        # --- bus ---
        self.aBus = Var(info='bus voltage angle',
                        unit='rad',
                        name='aBus', tex_name=r'a_{Bus}',
                        model='Bus', src='a',)

        # --- line flow ---
        self.plf = Var(info='Line flow',
                       unit='p.u.',
                       name='plf', tex_name=r'p_{lf}',
                       model='Line',)

    def unpack(self, res):
        """
        Unpack results from PYPOWER.
        """
        system = self.system
        mva = res['baseMVA']

        # --- copy results from ppc into system algeb ---
        # --- Bus ---
        system.Bus.v.v = res['bus'][:, 7]               # voltage magnitude
        system.Bus.a.v = res['bus'][:, 8] * deg2rad     # voltage angle

        # --- PV ---
        system.PV.p.v = res['gen'][system.Slack.n:, 1] / mva        # active power
        system.PV.q.v = res['gen'][system.Slack.n:, 2] / mva        # reactive power

        # --- Slack ---
        system.Slack.p.v = res['gen'][:system.Slack.n, 1] / mva     # active power
        system.Slack.q.v = res['gen'][:system.Slack.n, 2] / mva     # reactive power

        # --- Line ---
        self.plf.optz.value = res['branch'][:, 13] / mva  # line flow

        # --- copy results from system algeb into routine algeb ---
        for vname, var in self.vars.items():
            owner = getattr(system, var.model)  # instance of owner, Model or Group
            if var.src is None:          # skip if no source variable is specified
                continue
            elif hasattr(owner, 'group'):   # if owner is a Model instance
                grp = getattr(system, owner.group)
                idx = grp.get_all_idxes()
            elif hasattr(owner, 'get_idx'):  # if owner is a Group instance
                idx = owner.get_all_idxes()
            else:
                msg = f"Failed to find valid source variable `{owner.class_name}.{var.src}` for "
                msg += f"{self.class_name}.{vname}, skip unpacking."
                logger.warning(msg)
                continue
            try:
                logger.debug(f"Unpacking {vname} into {owner.class_name}.{var.src}.")
                var.optz.value = owner.get(src=var.src, attr='v', idx=idx)
            except AttributeError:
                logger.debug(f"Failed to unpack {vname} into {owner.class_name}.{var.src}.")
                continue
        self.system.recent = self.system.routines[self.class_name]
        return True

    def solve(self, method=None):
        """
        Solve DC power flow using PYPOWER.
        """
        ppc = system2ppc(self.system)
        ppopt = ppoption(PF_DC=True)

        res, sstats = runpf(casedata=ppc, ppopt=ppopt)
        return res, sstats

    def run(self, **kwargs):
        """
        Run DC pwoer flow.
        *args and **kwargs go to `self.solve()`, which are not used yet.

        Examples
        --------
        >>> ss = ams.load(ams.get_case('matpower/case14.m'))
        >>> ss.DCPF.run()

        Parameters
        ----------
        method : str
            Placeholder for future use.

        Returns
        -------
        exit_code : int
            Exit code of the routine.
        """
        if not self.initialized:
            self.init()
        t0, _ = elapsed()

        res, sstats = self.solve(**kwargs)
        self.converged = res['success']
        self.exit_code = 0 if res['success'] else 1
        _, s = elapsed(t0)
        self.exec_time = float(s.split(' ')[0])
        n_iter = int(sstats['num_iters'])
        n_iter_str = f"{n_iter} iterations " if n_iter > 1 else f"{n_iter} iteration "
        if self.exit_code == 0:
            msg = f"<{self.class_name}> solved in {s}, converged in "
            msg += n_iter_str + f"with {sstats['solver_name']}."
            logger.info(msg)
            try:
                self.unpack(res)
            except Exception as e:
                logger.error(f"Failed to unpack results from {self.class_name}.\n{e}")
                return False
            self.system.report()
            return True
        else:
            msg = f"{self.class_name} failed in "
            msg += f"{int(sstats['num_iters'])} iterations with "
            msg += f"{sstats['solver_name']}!"
            logger.warning(msg)
            return False

    def summary(self, **kwargs):
        """
        # TODO: Print power flow summary.
        """
        raise NotImplementedError

    def enable(self, name):
        raise NotImplementedError

    def disable(self, name):
        raise NotImplementedError



"""
Module for Model class.
"""

import logging
from collections import OrderedDict
from typing import Iterable

import numpy as np
from andes.core.common import Config
from andes.core.param import ExtParam
from andes.core.service import BaseService, BackRef
from andes.utils.func import list_flatten

from ams.core.documenter import Documenter
from ams.core.var import Algeb

from ams.utils.misc import deprec_get_idx

logger = logging.getLogger(__name__)


class Model:
    """
    Base class for power system scheduling models.

    This class is revised from ``andes.core.model.Model``.
    """

    def __init__(self, system=None, config=None):

        # --- Model ---
        self.system = system
        self.group = 'Undefined'

        self.algebs = OrderedDict()  # internal algebraic variables
        self.vars_decl_order = OrderedDict()  # variable in the order of declaration

        self.params_ext = OrderedDict()  # external parameters

        self.services = OrderedDict()  # service/temporary variables
        self.services_ref = OrderedDict()  # BackRefs
        self.services_sum = OrderedDict()  # VarSums

        self.config = Config(name=self.class_name)  # `config` that can be exported
        if config is not None:
            self.config.load(config)

        # basic configs
        self.config.add(OrderedDict((('allow_adjust', 1),
                                    ('adjust_lower', 0),
                                    ('adjust_upper', 1),
                                     )))
        self.docum = Documenter(self)

    def _all_vars(self):
        """
        An OrderedDict of States, ExtStates, Algebs, ExtAlgebs
        """
        return OrderedDict(list(self.algebs.items()))

    def _check_attribute(self, key, value):
        """
        Check the attribute pair for valid names while instantiating the class.

        This function assigns `owner` to the model itself, assigns the name and tex_name.
        """
        if isinstance(value, (Algeb, BaseService)):
            if not value.owner:
                value.owner = self
            if not value.name:
                value.name = key
            if not value.tex_name:
                value.tex_name = key
            if key in self.__dict__:
                logger.warning(f"{self.class_name}: redefinition of member <{key}>. Likely a modeling error.")

    def __setattr__(self, key, value):
        """
        Overload the setattr function to register attributes.

        Parameters
        ----------
        key : str
            name of the attribute
        value : [Algeb]
            value of the attribute
        """
        self._check_attribute(key, value)
        self._register_attribute(key, value)

        super(Model, self).__setattr__(key, value)

    def _register_attribute(self, key, value):
        """
        Register a pair of attributes to the model instance.

        Called within ``__setattr__``, this is where the magic happens.
        Subclass attributes are automatically registered based on the variable type.
        Block attributes will be exported and registered recursively.
        """
        if isinstance(value, Algeb):
            self.algebs[key] = value
        elif isinstance(value, ExtParam):
            self.params_ext[key] = value
        elif isinstance(value, BackRef):
            self.services_ref[key] = value
            self.services[key] = value

    @property
    def class_name(self):
        """
        Return the class name
        """
        return self.__class__.__name__

    def list2array(self):
        """
        Convert all the value attributes ``v`` to NumPy arrays.

        Value attribute arrays should remain in the same address afterwards.
        Namely, all assignments to value array should be operated in place (e.g., with [:]).
        """

        for instance in self.num_params.values():
            instance.to_array()

    def set_backref(self, name, from_idx, to_idx):
        """
        Helper function for setting idx-es to ``BackRef``.
        """

        if name not in self.services_ref:
            return

        uid = self.idx2uid(to_idx)
        self.services_ref[name].v[uid].append(from_idx)

    def get(self, src: str, idx, attr: str = 'v', allow_none=False, default=0.0):
        """
        Get the value of an attribute of a model property.

        The return value is ``self.<src>.<attr>[idx]``

        Parameters
        ----------
        src : str
            Name of the model property
        idx : str, int, float, array-like
            Indices of the devices
        attr : str, optional, default='v'
            The attribute of the property to get.
            ``v`` for values, ``a`` for address, and ``e`` for equation value.
        allow_none : bool
            True to allow None values in the indexer
        default : float
            If `allow_none` is true, the default value to use for None indexer.

        Returns
        -------
        array-like
            ``self.<src>.<attr>[idx]``

        """
        uid = self.idx2uid(idx)
        if isinstance(self.__dict__[src].__dict__[attr], list):
            if isinstance(uid, Iterable):
                if not allow_none and (uid is None or None in uid):
                    raise KeyError('None not allowed in uid/idx. Enable through '
                                   '`allow_none` and provide a `default` if needed.')
                return [self.__dict__[src].__dict__[attr][i] if i is not None else default
                        for i in uid]
        # FIXME: this seems to be an unexpected case originted from ANDES
        if isinstance(uid, Iterable):
            if None in uid:
                return [self.__dict__[src].__dict__[attr][i] if i is not None else default
                        for i in uid]
        return self.__dict__[src].__dict__[attr][uid]

    def set(self, src, idx, attr, value):
        """
        Set the value of an attribute of a model property.

        Performs ``self.<src>.<attr>[idx] = value``. This method will not modify
        the input values from the case file that have not been converted to the
        system base. As a result, changes applied by this method will not affect
        the dumped case file.

        To alter parameters and reflect it in the case file, use :meth:`alter`
        instead.

        Parameters
        ----------
        src : str
            Name of the model property
        idx : str, int, float, array-like
            Indices of the devices
        attr : str, optional, default='v'
            The internal attribute of the property to get.
            ``v`` for values, ``a`` for address, and ``e`` for equation value.
        value : array-like
            New values to be set

        Returns
        -------
        bool
            True when successful.
        """
        uid = self.idx2uid(idx)
        self.__dict__[src].__dict__[attr][uid] = value
        return True

    def alter(self, src, idx, value, attr='v'):
        """
        Alter values of input parameters or constant service.

        If the method operates on an input parameter, the new data should be in
        the same base as that in the input file. This function will convert
        ``value`` to per unit in the system base whenever necessary.

        The values for storing the input data, i.e., the parameter's ``vin``
        field, will be overwritten. As a result, altered values will be
        reflected in the dumped case file.

        Parameters
        ----------
        src : str
            The parameter name to alter
        idx : str, float, int
            The device to alter
        value : float
            The desired value
        attr : str
            The attribute to alter, default is 'v'.

        Notes
        -----
        New in version 0.9.14: Added the signature `attr` to alter specific attributes.
        This feature is useful when you need to manipulate parameter values in the system
        base and ensure that these changes are reflected in the dumped case file.
        """

        instance = self.__dict__[src]

        if hasattr(instance, 'vin') and (instance.vin is not None):
            uid = self.idx2uid(idx)
            if attr == 'vin':
                self.set(src, idx, 'vin', value / instance.pu_coeff[uid])
                self.set(src, idx, 'v', value=value)
            else:
                self.set(src, idx, 'vin', value)
                self.set(src, idx, 'v', value * instance.pu_coeff[uid])
        elif not hasattr(instance, 'vin') and attr == 'vin':
            logger.warning(f"{self.class_name}.{src} has no `vin` attribute, changing `v`.")
            self.set(src, idx, 'v', value)
        else:
            self.set(src, idx, attr=attr, value=value)

    def idx2uid(self, idx):
        """
        Convert idx to the 0-indexed unique index.

        Parameters
        ----------
        idx : array-like, numbers, or str
            idx of devices

        Returns
        -------
        list
            A list containing the unique indices of the devices
        """
        if idx is None:
            logger.debug("idx2uid returned None for idx None")
            return None
        if isinstance(idx, (float, int, str, np.integer, np.floating)):
            return self._one_idx2uid(idx)
        elif isinstance(idx, Iterable):
            if len(idx) > 0 and isinstance(idx[0], (list, np.ndarray)):
                idx = list_flatten(idx)
            return [self._one_idx2uid(i) if i is not None else None
                    for i in idx]
        else:
            raise NotImplementedError(f'Unknown idx type {type(idx)}')

    def _one_idx2uid(self, idx):
        """
        Helper function for checking if an idx exists and
        converting it to uid.
        """

        if idx not in self.uid:
            raise KeyError("<%s>: device not exist with idx=%s." %
                           (self.class_name, idx))

        return self.uid[idx]

    def doc(self, max_width=78, export='plain'):
        """
        Retrieve model documentation as a string.
        """
        return self.docum.get(max_width=max_width, export=export)

    @deprec_get_idx
    def get_idx(self):
        """
        Return the index of the model instance.
        Equivalent to ``self.idx.v``, develoepd for consistency with group method
        ``get_idx``.
        """
        return self.idx.v

    def get_all_idxes(self):
        """
        Return all the indexes of this model.

        .. note::
            New in version 1.0.0. Add to follow the group method ``get_all_idxes``.

        Returns
        -------
        list
            A list of indexes.
        """
        return self.idx.v

    def __repr__(self):
        dev_text = 'device' if self.n == 1 else 'devices'

        return f'{self.class_name} ({self.n} {dev_text}) at {hex(id(self))}'
